"======================================================================
|
|   Base class definition for C data type description objects.
| 
|   $Revision: 1.9 $
|   $Date: 1995/09/17 10:48:51 $
|   $Author: sbb $
|
 ======================================================================"



"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     16 Sep 95	  Removed separate valueType defintion for pointers
|			  and arrays; using the inherited version instead.
|
| sbb	      9 Sep 95	  Adjusted sizes of scalar types to use actual values
|			  for the platform instead of constant values that
|			  Smalltalk made up.
|
| sbb	      1 May 95	  Changed CType>>cObjectType to return CObject instead
|			  of subclass responsibility.  
|
| sbb	     29 Apr 95	  Moved the creation of the instances of CType
|			  subclasses into this file -- makes more sense here.
|
| sbb	     29 Apr 95	  Added cObjectType -- may be replacing what the
|			  baseType method was, but this is more descriptive.
|			  Also added a new operation on CType instances to
|			  create a new CObject subclass from the CType
|			  instance. 
|
| sbb	     29 Apr 95	  Removed baseType -- it appears to have no function
|			  now. 
|
| sbb	      2 Jan 95	  Hacked for new type architecture.
|
| sbb	      9 Apr 94	  Added alloc:type:
|
| sbb	     16 Feb 92	  created summer 90
|
"

Object subclass: #CType
       instanceVariableNames: 'cObjectType'
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

"Note: cObjectType is used to hold onto the CObject subclass that gets
 created for a given CType.  Used primarily in the C part of the interpreter
 because internally it cannot execute methods to get values, so it has a 
 simple to access instance variable which holds the desired subclass. "

CType comment: 'I am not part of the standard Smalltalk kernel class hierarchy.
I contain type information used by subclasses of CObject, which represents
external C data items.

My subclasses have instances which represent the actual data types; for the
scalar types, there is only one instance created of each, but for the 
aggregate types, there is at least one instance per base type and/or number of
elements.'!


!CType class methodsFor: 'instance creation'!

cObjectType: aCObjectClass
    ^self new init: aCObjectClass
! !



!CType methodsFor: 'C instance creation'!

new
    ^CObject alloc: self sizeof type: self
! !




!CType methodsFor: 'accessing'!

cObjectType
    ^cObjectType
! 

sizeof
    ^CPtrSize				"default value"
!

alignof
    ^CPtrSize
!

" valueType is used as a means to communicate to the interpreter the 
  underlying type of the data.  For anything but scalars, ptrs etc, it's just 
  'self' "
valueType
    ^self			"not a very good name for this"

! !


!CType methodsFor: 'private'!

init: aCObjectClass
    cObjectType _ aCObjectClass
! !





"!CType methodsFor: 'old-accessing'!

subType
    ^subType
!

baseType
    ^baseType
!

numElements
    ^numElements
!

scalarIndex
    ^baseType scalarIndex
!

sizeof
    ^numElements * subType sizeof
! !"


"Note that the type subclasses are CxxxCType, not CxxxType"

CType subclass: #CCharCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!CCharCType class methodsFor: 'instance creation'!

new
    ^super new init: CChar
! !


!CCharCType methodsFor: 'accessing'!

sizeof
    ^1
!

alignof
    ^1
!

valueType
    ^0
! !


CType subclass: #CUCharCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CUCharCType class methodsFor: 'instance creation'!

new
    ^super new init: CUChar
! !


!CUCharCType methodsFor: 'accessing'!

sizeof
    ^1
!

alignof
    ^1
!

valueType
    ^1
! !


CUCharCType subclass: #CByteCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

"Used to describe a type which is identical to CUCharCType except yields
 integer values when accessed instead of characters"

!CByteCType class methodsFor: 'instance creation'!

new
    ^super new init: CByte
! !



CType subclass: #CShortCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CShortCType class methodsFor: 'instance creation'!

new
    ^super new init: CShort
! !


!CShortCType methodsFor: 'accessing'!

sizeof
    ^2
!

alignof
    ^2
!

valueType
    ^2
! !


CType subclass: #CUShortCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CUShortCType class methodsFor: 'instance creation'!

new
    ^super new init: CUShort
! !


!CUShortCType methodsFor: 'accessing'!

sizeof
    ^2
!

alignof
    ^2
!

valueType
    ^3
! !


" !!! Port NOTE: DEC Alpha & other 64 bit architectures"

CType subclass: #CLongCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!CLongCType class methodsFor: 'instance creation'!

new
    ^super new init: CLong
! !


!CLongCType methodsFor: 'accessing'!

sizeof
    ^CLongSize
!

alignof
    ^CLongSize			"should be conditional"
!

valueType
    ^4
! !

" !!! Port NOTE: DEC Alpha & other 64 bit architectures"

CType subclass: #CULongCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!CULongCType class methodsFor: 'instance creation'!

new
    ^super new init: CULong
! !


!CULongCType methodsFor: 'accessing'!

sizeof
    ^CLongSize
!

alignof
    ^CLongSize			"should be conditional"
!

valueType
    ^5
! !


" !!! Port NOTE: DEC Alpha & other 64 bit architectures"

CType subclass: #CFloatCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!CFloatCType class methodsFor: 'instance creation'!

new
    ^super new init: CFloat
! !


!CFloatCType methodsFor: 'accessing'!

sizeof
    ^CFloatSize
!

alignof
    ^CFloatSize			"should be conditional"
!

valueType
    ^6
! !


CType subclass: #CDoubleCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CDoubleCType class methodsFor: 'instance creation'!

new
    ^super new init: CDouble
! !


!CDoubleCType methodsFor: 'accessing'!

sizeof
    ^CDoubleSize
!

alignof
    ^CDoubleAlignment
!

valueType
    ^7
! !




CType subclass: #CStringCType
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CStringCType class methodsFor: 'instance creation'!

new
    ^super new init: CString
! !

!CStringCType methodsFor: 'accessing'!

sizeof
    ^CPtrSize			"sizeof (char *)"
!

alignof
    ^CPtrSize			"alignof (char *)"
!

valueType
    ^8
! !




"Note: at some point, it may make sense to create a common super class for
CPtrCType and CArrayCType."

CType subclass: #CPtrCType
       instanceVariableNames: 'eltCType'
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 


!CPtrCType class methodsFor: 'instance creation'!

elementType: aCType
    ^self new init: aCType
! !


!CPtrCType methodsFor: 'accessing'!

sizeof
    ^CPtrSize			"size is the size of the pointer itself"
!

alignof
    ^CPtrSize			"alignment is for the pointer"
!

"16-Sep-95 22:44:59 letting this default to the inherited definition from 
 CType -- just return self.  What was happening was code would do 
 'elementType valeuType' which for a pointer to pointer to char was returning
 char, instead of pointer to char.  The comments for this method are
 out of date.
"
"valueType
    ^eltCType"			" this is probably correct, even though it"
				" looks odd -- the elementCType is a CType"
				" instance, which will be the next type of"
				" data that's needed "
"!"

elementType
    ^eltCType
! !


!CPtrCType methodsFor: 'private'!

init: aCType
    super init: CPtr.
    eltCType _ aCType.
! !



CType subclass: #CArrayCType
       instanceVariableNames: 'eltCType numElements'
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!CArrayCType class methodsFor: 'instance creation'!

elementType: aCType numberOfElements: anInteger
    ^self new init: aCType numberOfElements: anInteger
! !


!CArrayCType methodsFor: 'accessing'!

sizeof
    ^eltCType sizeof * numElements
!

alignof
    ^eltCType alignof
!

"16-Sep-95 22:47:45 see comments in CPtrCType>>valueType.

valueType
    ^eltCType
!"

numberOfElements
    ^numElements
! 

elementType
    ^eltCType
! !

!CArrayCType methodsFor: 'private'!

init: aCType numberOfElements: anInteger
    super init: CArray.
    eltCType _ aCType.
    numElements _ anInteger.
! !


"This may be confusing -- the *CType objects are subclasses of CType -- which
 are C type descriptors.  The C*Type objects are instances of those CType
 subclasses and are only really used as an abbreviation for creating an
 instance of the type, and reusing a common instance. "

Smalltalk at: #CCharType    put: CCharCType new.
Smalltalk at: #CUCharType   put: CUCharCType new.
Smalltalk at: #CShortType   put: CShortCType new.
Smalltalk at: #CUShortType  put: CUShortCType new.
Smalltalk at: #CLongType    put: CLongCType new.
Smalltalk at: #CULongType   put: CULongCType new.
Smalltalk at: #CFloatType   put: CFloatCType new.
Smalltalk at: #CDoubleType  put: CDoubleCType new.
Smalltalk at: #CStringType  put: CStringCType new.
Smalltalk at: #CByteType    put: CByteCType new
!
