"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     17 Feb 92	  Created Summer 90.
|
"

Object subclass: #Host
       instanceVariableNames: 'hostEnt hostName hostAliases addrType 
    	    addrLength addrList'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'Hacking' !

Host comment: 'I am a class for accessing Internet host information.'!

Behavior defineCFunc: 'gethostent'
	 withSelectorArgs: 'getHostEnt'
	 forClass: Host class
	 returning:#cObject
	 args: #()
!

Behavior defineCFunc: 'gethostbyname'
	 withSelectorArgs: 'getHostByName: aString'
	 forClass: Host class
	 returning: #cObject
	 args: #(string)
!

Behavior defineCFunc: 'getname'
	 withSelectorArgs: 'getName: aHostEnt'
	 forClass: Host
	 returning: #string
	 args: #(cObject)
!

Behavior defineCFunc: 'getalias'
	 withSelectorArgs: 'getAlias: aHostEnt at: anIndex'
	 forClass: Host
	 returning: #string
	 args: #(cObject int)
!

Behavior defineCFunc: 'getaddrtype'
	 withSelectorArgs: 'getAddrType: aHostEnt'
	 forClass: Host
	 returning: #int
	 args: #(cObject)
!

Behavior defineCFunc: 'gethlength'
	 withSelectorArgs: 'addrLengthOf: aHostEnt'
	 forClass: Host
	 returning: #int
	 args: #(cObject)
!

Behavior defineCFunc: 'getaddrlist'
	 withSelectorArgs: 'getAddrList: aHostEnt at: anIndex into: byteArray'
	 forClass: Host
	 returning: #int
	 args: #(cObject int byteArrayOut)
!

!Host class methodsFor: 'instance creation'!

forHost: aHostName
    ^self new init: (self getHostByName: aHostName)
!

nextHost
    ^self new init: (self getHostEnt)
!!

!Host methodsFor: 'accessing'!

hostName
    ^hostName
!

hostAlias: anIndex
    ^hostAliases at: anIndex
!

hostAliases
    ^hostAliases
!

hostAddrType
    ^addrType
!

hostAddrList: anIndex
    ^addrList at: anIndex
!!


!Host methodsFor: 'private'!

init: aCObject
    | i aliases alias addresses bytes |
    hostEnt _ aCObject.
    hostName _ self getName: hostEnt.
    i _ 0.
    aliases _ OrderedCollection new.
    [ alias _ self getAlias: hostEnt at: i.
      alias notNil ] whileTrue:
	  [ aliases add: alias.
	    i _ i + 1 ].
    hostAliases _ aliases "asArray". 	"Don't know if asArray is good"
    addrType _ self getAddrType: hostEnt.
    addrLength _ self addrLengthOf: hostEnt.
    i _ 0.
    addresses _ OrderedCollection new.
    [ bytes _ self hostAddrListAt: i.
      (bytes at: 1) ~= 0 ] whileTrue:
	  [ addresses add: bytes.
	    i _ i + 1 ].
    addrList _ addresses
!

hostAddrListAt: anIndex
    | bytes |     
    bytes _ ByteArray new: addrLength.
    self getAddrList: hostEnt at: anIndex into: bytes.
    ^bytes
!!
