"File CtxtInspect.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date            Change
| bdiller       7 July 95.      Derived from sbb's BBrowser.st
|
"
! Object methodsFor: 'debugging' !

inspect: pane
"Open an Inspector window on self"
    Inspector new openOn: self in: pane
!!

! Dictionary methodsFor: 'debugging' !

inspect: pane
"Open a DictionaryInspector window on self"
    DictionaryInspector new openOn: self in: pane
!!

GuiData subclass: #ContextInspector
     instanceVariableNames: 'contextHolder theClass theSelector'
     classVariableNames: ''
     poolDictionaries: ''
     category: nil
! 

!ContextInspector class methodsFor: 'instance creation'!

new: contextBrowser
    ^self new init: contextBrowser
! !

!ContextInspector methodsFor: 'private'!

init: contextBrowser
    | topView toplevel container text lowerpane pane list context llpane lrpane textView|
    contextHolder _ contextBrowser.
    context _ contextHolder currentContext.

    theSelector _ context selector.
    theSelector isNil ifTrue: ['Unable to inspect method context' printNl. ^self].

    topView _ ((TopLevelShell new: 'Context Inspector' ) data: self).
    toplevel _ topView blox.

		self setBackgroundColorOn: topView blox.

    pane := Form new: 'panes' in: topView.
    topView addChildView: pane.
    
		self setBackgroundColorOn: pane blox.
    pane addChildView: ((textView := PText new: pane)
    data: self;
    stateChange: #text;
    handleUserChange: #compile:from:;
    textMsg: #text;
    yourself).
    textView menuInit: ((PopupMenu new: textView label: '') selectors: 
   #(('cut' gstCut) ('copy' gstCopy) ('paste' gstPaste) 
    () ('clear' gstClear) () ('line...' line)  ('find...' find) ()
    ('do it' eval) ('print it' evalAndPrintResult) ('inspect' evalAndInspectResult) 
    () ('accept' compileIt)
    ('cancel' revert) () ('close' close))
                receiver: textView
                argument: nil).
    text _ textView blox.
    text width: 648; height: 200.
    
    theClass _ (context receiver class includesSelector: theSelector)
        ifTrue: [ context receiver class]
        ifFalse: [  context receiver class 
                        whichClassIncludesSelector: theSelector ].

    text contents: (theClass sourceCodeAt: theSelector).

    lowerpane := Form new: 'panes' in: pane.
    pane addChildView: lowerpane.
    self setBackgroundColorOn: pane blox.
    lowerpane blox posVert: (textView blox).
        
    llpane := Form new: 'panes' in: lowerpane.
    lowerpane addChildView: llpane.
    self setBackgroundColorOn: llpane blox.

    contextHolder currentContext receiver inspect: llpane.

    lrpane := Form new: 'panes' in: lowerpane.
    lowerpane addChildView: lrpane.
    self setBackgroundColorOn: lrpane blox.

    StackInspector new openOn: (contextHolder currentContext) in: lrpane.
    lrpane blox posHoriz: (llpane blox).

    toplevel map.
!!

!ContextInspector methodsFor: 'text pane'!

text
"Return source code for the selected method"
    theSelector notNil 
        ifTrue: [^theClass sourceCodeAt: theSelector].
    ^''.
!

compile: aString from: aView
"Compile aString derived from text in text view for the selected selector"
    theSelector notNil  
        ifTrue: [ theClass compile: aString classified: 
                      ((theClass compiledMethodAt: theSelector) methodCategory) 
                           ifError: [:fname :lineNo :errorString |
                                                                  aView displayError: errorString at: lineNo.
                                                                  ^nil]]
!!

!ContextInspector methodsFor: 'private'!

setBackgroundColorOn: win
    | color |
"create a nice purple"
    color _ Blox createColor: 16rFFFF green: 0 blue: 16rFFFF.
    win backgroundColor: color.
! !















