"File MethSetBrow.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date            Change
| bdiller       11 Oct 93       Created.
"
GuiData subclass:  #MethodSetBrowser
        instanceVariableNames: 'methodList theClass theSelector selection '
        classVariableNames: ''
        poolDictionaries: ''
        category: 'Graphics Windows'
!

!MethodSetBrowser methodsFor: 'initializing'!

openOn: aMethodDictionary title: name  selection: aSymbol
"Open a method set browser.  The argument aMethodDictionary consists of 
alpha-sorted collection of strings.  Each element is of two forms: 1) className 
class selector, or 2) className selector.  This browser consists of two 
vertically placed panes.  The top pane is a list which displays the sorted 
methods in aMethodDictionary.  The bottom pane is a text view which will 
display the source code for a selector which is selected from the top pane.  
        In general, aSymbol denotes a selector.  If this parameter is non-nil, the 
first occurence of aSymbol will be selected in the text view when a 
selector is first selected from the top pane" 
    | topView childView aStream listView textView container | 
    aSymbol notNil 
	"Parse selector expression, aSymbol, inclusive of first colon"
	ifTrue:[aStream := WriteStream on: (String new: 0). 
		aSymbol detect: [ :ch | aStream nextPut: ch. ch == $:]
			ifNone: [0].
		selection := aStream contents].
    topView := (TopLevelShell new: 'MethodSetBrowser') data: self.
    topView blox x: 0.
    topView blox y: 330.
		self setBackgroundColorOn: topView blox.

    topView blox height: 316.
    topView blox width: 408.
    "Use PanedWindow class to manage the list and text view panes"
    childView := Form new: 'Form' in: topView.
    topView addChildView: childView.
    container := childView blox.      
    self setBackgroundColorOn: container.
    "Create a list in top half of window"
    childView addChildView: ((listView := PList new: 'MethodSet' in: childView)
				 initialize;
				 data: self;
				 stateChange: #methodList; 
				 handleUserChange: #methodSelection:; 
				 listMsg: #methodList;
				 yourself).
    listView blox height: 150.
    listView blox width: 400.
    "Create a text view and install in lower half of window"
    childView addChildView: ((textView := PText new: childView)
				 data: self;
				 stateChange: #text;
				 handleUserChange: #compile:from:;
				 textMsg: #text;
				 selection: selection;
				 yourself).
    textView menuInit: ((PopupMenu new: textView label: '') selectors: 
                     #(('cut' gstCut) ('copy' gstCopy) ('paste' gstPaste) 
		     () ('clear' gstClear) () ('line...' line)  ('find...' find) ()
                     ('do it' eval) ('print it' evalAndPrintResult) ('inspect' evalAndInspectResult)
                     () ('accept' compileIt) ('cancel' revert) () ('close' close))
							    receiver: textView
							    argument: nil).
    textView blox height: 150.
    textView blox width: 400.
    textView blox posVert: (listView blox).
    "Initialize instance variable, methodList, which governs list display"
	
     methodList := aMethodDictionary.
     self changeState: #methodList.
     Primitive updateViews.
    "Initialize all the manufactured widgets"
    topView display
!!

!MethodSetBrowser methodsFor: 'accessing'!

methodList
    ^methodList
!

methodSelection: assoc
"Derive class and selector from list selection.  The selection is derived
from an item in the method list pane.  A list item may be of two
forms: 1) className class selector, or 2) className selector. Form (1)
contains 3 string tokens and form (2) contains 2.  To derive the class
from form (1), the instance class is derived from the Smallltalk
dictionary using the first string token as a key.  Then class is sent
to the instance class to derive the class of the instance class.  The
selector is derived from the third token.  In form (2), the instance
class is derived directly from the first string token.  The selector
is obtained from the second token"
    | parsedString |
    assoc value isNil ifTrue: [^theSelector := nil].
    parsedString := (TokenStream on: assoc value) contents.
    parsedString size = 3 
	ifTrue: [theClass := (Smalltalk at: (parsedString at: 1) asSymbol) class.
		 theSelector :=  (parsedString at: 3) asSymbol]
	ifFalse: [ theClass := Smalltalk at: (parsedString at: 1) asSymbol.
		   theSelector :=  (parsedString at: 2) asSymbol].
    self changeState: #text.
    Primitive updateViews.
!

text
"Return source code for the selected method"
    theSelector notNil 
	ifTrue: [^theClass sourceCodeAt: theSelector].
    ^''.
!!


!MethodSetBrowser methodsFor: 'text view blue button menu'!

compile: aString from: aView
"Compile aString derived from text in text view for the selected selector"
    theSelector isNil ifTrue: [^aView beep].
	  theClass compile: aString classified: 
	      ((theClass compiledMethodAt: theSelector) methodCategory) 
			   ifError: [:fname :lineNo :errorString |
				  aView displayError: errorString at: lineNo.
  				  ^nil ].
!

selection
    ^selection
!!

!MethodSetBrowser methodsFor: 'private'!

setBackgroundColorOn: win
    | color |
"create a nice green"
    color _ Blox createColor: 16r8000 green: 16rFFFF blue: 16r8000.
    win backgroundColor: color.
! !

