
"
| This is a test app for the interface between smalltalk and objective that
| lets the user browser smalltalk code inspecting various things and maybe
| even allowing people to alter the code and save it.
|
| Classes --
|
|	STNormalBrowserDelegate	
|		* ClassBrowser
|		* MethodBrowser
|
|	STAppDelegate
|		* AppDelegate
|
|	STObject
|		* BrowserController
|		* WorkspaceController
|		* ConsoleController
| ============================================================================
|
"

"
| ============================================================================
| If there is a method that wants to return a string by reference you need to
| find out the OCInterface class method performs the actual callout and 
| replace the string value with a stringOut or the string will never be update.
| Make sure that the string you passin is large enough to hold the value that
| will be placed in it or the system will crash. 
| ============================================================================
"

Behavior defineCFunc: 'msgSend'
		withSelectorArgs: 'intMsgSend3: anId sel: aSelector argstring: arg0 argint: arg1 argint: arg2'
		forClass: OCInterface class
		returning: #int
		args: #(cObject cObject stringOut int int)
!

"
| ============================================================================
| ============================================================================
"

!STObject methodsFor: 'Fixing'!

setOOP: arg0 
        | anId aSelector rValue | 
        anId _ id.
        aSelector _ OCInterface getUid: 'setOOP:'.

		Behavior defineCFunc: 'msgSend'
			withSelectorArgs: 'cObjectMsgSend1: anId sel: aSelector argSmalltalk: arg0'
			forClass: OCInterface class
			returning: #cObject
			args: #(cObject cObject smalltalk).
			
        rValue _ OCInterface cObjectMsgSend1: anId sel: aSelector argSmalltalk: arg0.
        ((OCInterface id: id equalTo: rValue) = 1)
                ifTrue: [ ^self ]
                ifFalse: [ ^OCObject newFromId: rValue. ].
!

init
	super init.
	self setOOP: self.
!!

"
| ============================================================================
| ============================================================================
"

!ClassDescription methodsFor: 'filing'!

fileOutCategory: categoryName on: aFileStream
    self emitCategory: categoryName toStream: aFileStream.
!!

!Behavior methodsFor: 'Category Info'!

categories
	" Returns a set with all the categories we have "
	| methodDict categories |
	categories _ Set new.
	methodDict _ self methodDictionary.
	methodDict notNil
		ifTrue: [ methodDict do: [ :aMethod | categories add: (aMethod methodCategory) ]. ].
	^categories
!

selectorsIn: aCategory
	" Returns a set of all selectors in a category of self "
	| methodDict selectors selectorsInCategory |
	selectorsInCategory _ Set new.
	methodDict _ self methodDictionary.
	selectors _ self selectors.
	
	methodDict notNil
		ifTrue: [
			selectors do: [ :sel | 					
				(((methodDict at: sel) methodCategory) = aCategory) 
					ifTrue: [ selectorsInCategory add: sel ] ] ] .
	^selectorsInCategory
!!

"
| ============================================================================
| ============================================================================
"

!Text methodsFor: 'Appending'!

appendText: aString
	| len |
	len _ self textLength.
	len == 0
		ifTrue: [ self setText: aString ]
		ifFalse: [ self setSel: len anArg: len; replaceSel: aString ].
	self scrollSelToVisible.
!!

"
| ============================================================================
| Below are some things that appear as defines that are used regulary in
| NS
| ============================================================================
"

Smalltalk at:#YES put: (1 asCharacter)!
Smalltalk at:#NO put: (0 asCharacter)!

"
| ============================================================================
| Some forward defines
| ============================================================================
"

Smalltalk at:#MethodBrowserController put: nil!

"
| ============================================================================
| Subclass of ScrollView with text inside setup as it should be
| ============================================================================
"

ScrollView subclass: #TextScrollView
		 instanceVariableNames: 'text'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

TextScrollView comment: 'This is an attempt to mimmick the functionality of the
text scrollview found in Interface Builder.' !

!TextScrollView methodsFor: 'Initialization'!

initFrame: aRect
	| size |

	size _ Size alloc initFromWidth: (aRect width) height: (aRect height).

	super initFrame: (aRect asCPointer).
	
	self setOpaque: YES.
	self notifyAncestorWhenFrameChanged: YES.
	self setAutoresizeSubviews: YES.
	self setBorderType: 2.
	self setOpaque: YES.
	self setBackgroundGray: 1.0.
	self setVertScrollerRequired: YES.
	self setHorizScrollerRequired: NO .
	self setDynamicScrolling: YES.
	
	text _ Text alloc initFrame: (aRect asCPointer).
	text setMonoFont: NO.
    text setOpaque: YES.	
	text notifyAncestorWhenFrameChanged: YES.
	text setAutoresizeSubviews: YES.
	text setVertResizable: YES.
	text setHorizResizable: YES.
	text setMinSize: (size asCPointer).
	size setHeight: 100000000.0.
	size setWidth: 10000000.0.
	text setMaxSize: (size asCPointer).
	text setAutosizing: (2 bitOr: 16).
	
	text setSelectable: YES.
	text setEditable: YES.
	self setDocView: (text id).
	self display.
!!

!TextScrollView methodsFor: 'Accessing'!

text
	^text
!!

"
| ============================================================================
| Below is the ClassBrowserDelegate class code
| ============================================================================
"


STNormalBrowserDelegate subclass: #ClassBrowserDelegate
		 instanceVariableNames: ''
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

ClassBrowserDelegate comment: 'Delegate to a NXBrowser that provides the browser
will data to browse a class heirarchy'!

!ClassBrowserDelegate methodsFor: 'Browsing'!

browser: arg0 titleOfColumn: arg1 
	^'Class Heirarchy'
!

browser: arg0 fillMatrix: arg1 inColumn: arg2
	| sender matrix column cell selString selClass classes i sorted |
	matrix _ Matrix newWithId: arg1.
	sender _ NXBrowser newWithId: arg0.
	column _ arg2.
	
	(column == 0)
		ifTrue: [
			matrix addRow.
			cell _ matrix cellAt: 0 anArg: 0.
			cell setStringValue: 'Object'.
			cell setLoaded: YES.
			cell setLeaf: NO. 
			^1.]
		ifFalse: [
			cell _ sender selectedCell.
			selClass _ Smalltalk at: ((cell stringValue) asSymbol).
			classes _ selClass subclasses.  
			sorted _ Set new.
			classes do: [ :class | sorted add: (class name) ].
			
			i _ 0.			
			sorted asSortedCollection
				do: [ :element | 
					matrix addRow.
					cell _ matrix cellAt: i anArg: 0.
					cell setStringValue: element.
					
					((Smalltalk at: (cell stringValue asSymbol)) isKindOf: (OCObject class))
						ifTrue: [ cell setFont: ((Font newFont: 'Helvetica-Oblique' size: 12.0) id). ].
					((Smalltalk at: (cell stringValue asSymbol)) isKindOf: (STObject class))
						ifTrue: [ cell setFont: ((Font newFont: 'Helvetica-BoldOblique' size: 12.0) id). ].

					cell setLoaded: YES.
						
					((Smalltalk at: (cell stringValue asSymbol)) subclasses size) > 0 
						ifTrue: [ cell setLeaf: NO ]
						ifFalse: [ cell setLeaf: YES ].
 					i _ i + 1. ].				
			^i ].
	^0			
!

browser: arg0 getNumRowsInColumn: arg1
	^0
!!

"
| ============================================================================
| Below is the MethodBrowserDelegate class code
| ============================================================================
"

STNormalBrowserDelegate subclass: #MethodBrowserDelegate
		 instanceVariableNames: 'isInstBrowsing curClass'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

MethodBrowserDelegate comment: 'Delegate to a browser that display the categories
and methods for a specific class' !

!MethodBrowserDelegate methodsFor: 'Instance initialization'!

init
	"Inits instance vars"
	super init.
	isInstBrowsing _ true.
	curClass _ Object.
!!

!MethodBrowserDelegate methodsFor: 'Browsing'!

browseInstanceMethods
	" Sets isInstBrowsing to true "
	isInstBrowsing _ true.
!

browseClassMethods
	" Sets isInstBrowsing to false "
	isInstBrowsing _ false.
!

browseClass: aClass
	" Sets isInstBrowsing to true "
	curClass _ aClass.
!

	
browser: arg0 titleOfColumn: arg1 
	isInstBrowsing	
		ifTrue: [ ^'Instance Categories' ]
		ifFalse: [ ^'Class Categories' ]
!

browser: arg0 fillMatrix: arg1 inColumn: arg2
	| sender matrix column cell categories i catSelectors curCategory |
	matrix _ Matrix newWithId: arg1.
	sender _ NXBrowser newWithId: arg0.
	column _ arg2.

	(column == 0)
		ifTrue: [
			" Find out how many categories we have "

			isInstBrowsing
				ifTrue: [ categories _ curClass categories. ]
				ifFalse: [ categories _ curClass class categories. ].

			i _ 0.
			((categories size) == 0)
				ifTrue: [ ^0 ].						
			categories asSortedCollection
				do: [ :cat |
					matrix addRow.
					cell _ matrix cellAt: i anArg: 0.
					cell setStringValue: (cat asString).
					cell setLoaded: YES.
					cell setLeaf: NO.
					i _ i + 1 ]. 
				^i. ]
		ifFalse: [
			curCategory _ sender selectedCell stringValue.

			isInstBrowsing
				ifTrue: [ catSelectors _ curClass selectorsIn: curCategory. ]
				ifFalse: [ catSelectors _ curClass class selectorsIn: curCategory. ].

			i _ 0.
			catSelectors asSortedCollection
				 do: [ :sel |
					matrix addRow.
					cell _ matrix cellAt: i anArg: 0.
					cell setStringValue: (sel asString).
					cell setLoaded: YES.
					cell setLeaf: YES.
					i _ i + 1. ].
				^i ].
	^0
!

browser: arg0 getNumRowsInColumn: arg1
	^0
!!

"
|
| ============================================================================
|
"

STObject subclass: #ClassEditor
		 instanceVariableNames: 'panel scrolledText fieldMatrix actionMatrix'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

ClassEditor comment: 'I control a simple panel that popups when a class is edited
or being create.  I then update the system with the new class info.' !

!ClassEditor methodsFor: 'Initialization'!

init
	| rect |
	super init.
	rect _ Rect alloc initFromX: 500.0 y: 600.0 width: 400.0 height: 200.0.
	self initPanel: rect.
	self initText: rect.
	self initActionMatrix: rect.
	self compose.
!

initPanel: rect
	panel _ Panel alloc initContent: (rect asCPointer)
				style: 1
				backing: 2
				buttonMask: 1
				defer: YES.
	panel setTitle: 'Class Editor'.
!

initText: aRect
	| bounds font |
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	bounds setX: 10.0; setY: 40.0; setHeight: ((aRect height) - 45.0); setWidth: ((aRect width) - 20.0).
	scrolledText _ TextScrollView alloc initFrame: bounds.
	font _ Font newFont: 'Ohlfs' size: 10.0.
	(scrolledText text) setFont: (font id).
!

initActionMatrix: rect
	| bounds size |
	
	bounds _ Rect alloc initFromRect: (rect asCPointer).
	size _ Size alloc init.
	bounds setX: ((rect width) - 140.0); setY: 5.0 ; setHeight: 28.0 ; setWidth: 120.0.	
	
	actionMatrix _ Matrix alloc initFrame: (bounds asCPointer)
							mode: 0							"NX_LISTMODE"
							cellClass: (ButtonCell isa)
							numRows: 1
							numCols: 2.
	actionMatrix setAutosizeCells: YES.
	actionMatrix setIntercell: (size asCPointer).
	actionMatrix setTitle: 'Cancel' at: 0 anArg: 0.
	actionMatrix setTitle: 'Ok' at: 0 anArg: 1.
!

compose
	| contentView |
	contentView _ panel contentView.
	contentView notifyAncestorWhenFrameChanged: YES.
	contentView setAutoresizeSubviews: YES.

	"contentView addSubview: fieldMatrix id."
	contentView addSubview: scrolledText id.	
	contentView addSubview: actionMatrix id.
!!

!ClassEditor methodsFor: 'Display'!

orderFront
	panel makeKeyAndOrderFront: panel id.
	panel contentView display.
!!

!ClassEditor methodsFor: 'Class Editing'!

editClass: aClass
    | stream class name supername|
    
    aClass isNil 
		ifTrue: [ ^nil ].
    
	class _ aClass.

    supername _ class superclass name.		"Handle the nil case"
    name _ class name.						"handle the nil case"

    stream _ WriteStream on: (String new: 50).
    stream nextPutAll: supername;
		space;
		nextPutAll: class kindOfSubclass;
		nextPut: $#;
		nextPutAll: name; 
		nl.

    stream nextPutAll: '    instanceVariableNames: '''.
	class instVarNames do: [ :name | stream nextPutAll: name; space. ].
    stream nextPut: $'; nl.
    stream nextPutAll: '    classVariableNames: '''.
    class classVarNames do: [ :name | stream nextPutAll: name; space. ].
    stream nextPut: $'; nl.
    stream nextPutAll: '    poolDictionaries: '''.
    stream nextPut: $'; nl.
    stream nextPutAll: '    category: '''.
    stream print: class category; nextPut: $'; nl.
	scrolledText text setText: (stream contents).
	self orderFront
!!

!ClassEditor methodsFor: 'Class Creating'!

createSubclassOf: aClass
	| stream |
    stream _ WriteStream on: (String new: 0).
	stream nextPutAll:'
 ';
			nextPutAll: (aClass name), ' subclass: My', (aClass name);
			nextPutAll: '
	instanceVariableNames: '' ''
	classVariableNames: '' ''
	poolDictionaries: '' ''
	category: '' '''.  
	scrolledText text setText: (stream contents).
	self orderFront
!!

"
|
| ============================================================================
|
"

STObject subclass: #ClassBrowserController
		 instanceVariableNames: 'scrolledText classBrowser splitView box window mainMenu
		 						 app classBrowserDelegate buttonMatrix classEditor'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

ClassBrowserController comment: 'Coordinates the actions of all the browsers and the
text view which displays code and other junk' !

!ClassBrowserController methodsFor: 'Initialization'!

init
	| rect |
	super init.
	rect _ Rect alloc initFromX: 400.0 y: 600.0 width: 500.0 height: 400.0.
	self initMainWindow: rect.
	self initClassBrowser: rect.
	self initText: rect.
	self initButtonMatrix: rect.
	self initSplitView: rect.
	self compose.
!!

!ClassBrowserController methodsFor: 'Creating The Interface'!

compose
	| contentView |
	" Arranges everything on screen so that it looks nice "
	
	contentView _ window contentView.
	contentView notifyAncestorWhenFrameChanged: YES.
	contentView setAutoresizeSubviews: YES.

	splitView addSubview: classBrowser id.
	splitView addSubview: buttonMatrix id.
	splitView addSubview: scrolledText id.

	contentView addSubview: box id.
	contentView display.
!

initMainWindow: aRect
	window _ Window alloc initContent: (aRect asCPointer)
				style: 6
				backing: 2
				buttonMask: (4 bitOr: 1)
				defer: YES.
	window setTitle: 'Class Browser'.
	window makeKeyAndOrderFront: (window id).		
!

initButtonMatrix: aRect
	| bounds size |
	
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	size _ Size alloc init.
	bounds setX: 0.0 ; setY: 0.0 ; setHeight: 60.0.
	
	buttonMatrix _ Matrix alloc initFrame: (bounds asCPointer)
							mode: 0							"NX_LISTMODE"
							cellClass: (ButtonCell isa)
							numRows: 1
							numCols: 3.
	"buttonMatrix setAutosizing: (1 bitOr: 8)."
	buttonMatrix setAutosizeCells: YES.
	buttonMatrix setIntercell: (size asCPointer).
	
	buttonMatrix setTitle: 'Subclass...' at: 0 anArg: 0.
	buttonMatrix setTitle: 'Edit...' at: 0 anArg: 1.
	buttonMatrix setTitle: 'File Out...' at: 0 anArg: 1.
	
	buttonMatrix setIcon: 'AddClass.tiff' at: 0 anArg: 0.
	buttonMatrix setIcon: 'EditClass.tiff' at: 0 anArg: 1.
	buttonMatrix setIcon: 'FileOut.tiff' at: 0 anArg: 2.
	
	buttonMatrix setTag: 0 at: 0 anArg: 0.
	buttonMatrix setTag: 1 at: 0 anArg: 1.
	buttonMatrix setTag: 2 at: 0 anArg: 2.
	
	buttonMatrix setAction: (OCInterface getUid: 'action2:').
	buttonMatrix setTarget: self id.
!

initSplitView: aRect
	| bounds contentView size |
	bounds _ Rect alloc initFromRect: (aRect asCPointer).	
	bounds setX: 0.0 ; setY: 0.0.
	
	box _ Box alloc initFrame: (bounds asCPointer).
	box notifyAncestorWhenFrameChanged: YES.
	box setAutoresizeSubviews: YES.
	box setAutosizing: (2 bitOr:8; bitOr: 16).
	box setTitlePosition: 0.						" 0 = No Title "
	box setBorderType: 0.							" 0 = No Border "
	
	splitView _ NXSplitView alloc initFrame: (bounds asCPointer).
	splitView notifyAncestorWhenFrameChanged: YES.
	splitView setAutoresizeSubviews: YES.
	splitView setAutosizing: (2 bitOr:8; bitOr: 16).
	
	box setContentView: (splitView id).
!

initClassBrowser: aRect
	| bounds prototype |
	
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	bounds setX: 0.0 ; setY: 68.0 ; setHeight: (((aRect height) / 2.0) - 30.0).
	
	prototype _ NXBrowserCell alloc initTextCell: 'Browser'.
	prototype setFont: ((Font newFont: 'Helvetica' size: 12.0) id).
	
	classBrowser _ NXBrowser alloc initFrame: (bounds asCPointer).
	classBrowser setAutosizing: (2 bitOr: 16).
	classBrowserDelegate _ ClassBrowserDelegate alloc init.
	classBrowser setCellPrototype: prototype id.
	classBrowser setDelegate: (classBrowserDelegate id).
	classBrowser setTarget: (self id).
	classBrowser setAction: (OCInterface getUid: 'action1:').
	classBrowser setDoubleAction: (OCInterface getUid: 'doubleAction1:').
	classBrowser setEmptySelectionEnabled: YES.
	classBrowser setHorizontalScrollButtonsEnabled: NO.
	classBrowser setHorizontalScrollerEnabled: YES.
	classBrowser setMaxVisibleColumns: 3.
	classBrowser loadColumnZero.
!

initText: aRect
	| bounds font |
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	bounds setHeight: (((aRect height) / 2.0) - 30.0).
	scrolledText _ TextScrollView alloc initFrame: bounds.
	font _ Font newFont: 'Ohlfs' size: 10.0.
	(scrolledText text) setFont: (font id).
!!

!ClassBrowserController methodsFor: 'Target Actions'!

action1: sender
	| cell selectedClass stream class text |
	
	text _ scrolledText text.
	cell _ classBrowser selectedCell.
	selectedClass _ Smalltalk at: (cell stringValue asSymbol).

	stream _ WriteStream on: (String new: 20).
	stream nextPutAll: '
	Instance Variables: ';
		nextPutAll: selectedClass instanceVariableString;
		nextPutAll:'
		
';
"
		nextPutAll: '
	Class Variables: ';
		nextPutAll: ((selectedClass class) classVarNames);
"
		nextPutAll:'
		
';
	 	nextPutAll: (selectedClass comment).

	text setText: stream contents.
	text sizeToFit.
	text scrollSelToVisible.
	scrolledText display.
!

action2: sender
	" Popup a window for user to enter values for a new class"
	| matrix tag |
	matrix _ Matrix newFromId: sender.
	tag _ matrix selectedCell tag.
	tag == 0
		ifTrue: [ self createClass. ].
	tag == 1
		ifTrue: [ self editClass. ].
	tag == 2
		ifTrue: [ self fileOutClass. ].
!

editClass
	| cell class |
	cell _ classBrowser selectedCell.
	class _ Smalltalk at: (cell stringValue asSymbol).

	classEditor isNil
		ifTrue: [ classEditor _ ClassEditor alloc init ].
	classEditor editClass: class.
!

createClass
	| cell class |
	cell _ classBrowser selectedCell.
	class _ Smalltalk at: (cell stringValue asSymbol).

	classEditor isNil
		ifTrue: [ classEditor _ ClassEditor alloc init ].
	classEditor createSubclassOf: class.
!

fileOutClass
	| cell category savePanel value file class rect filename dir |
	cell _ classBrowser selectedCell.
	class _ Smalltalk at: (cell stringValue asSymbol).
	
	cell notNil
		ifTrue: [
			rect _ Rect alloc initFromX: 200.0 y: 200.0 width: 500.0 height: 400.0.
			savePanel _ SavePanel newContent: (rect asCPointer) style: 6 backing: 2 buttonMask: (4 bitOr: 1) defer: YES.
			savePanel setRequiredFileType: 'st'.
			savePanel setTitle: 'Filing Out'.
			value _ savePanel runModal. 
			
			value == 1
				ifTrue: [
					filename _ savePanel filename.
					dir _ savePanel directory.
					file _ FileStream open: filename mode: 'w'.
					class fileOutOn: file.
					file close ]. ].
!

doubleAction1: sender
	| cell class mBrowser |
	cell _ classBrowser selectedCell.
	class _ Smalltalk at: (cell stringValue asSymbol).
	mBrowser _ MethodBrowserController alloc initWithClass: class.
	(mBrowser window) makeKeyAndOrderFront: nil.
!!

"
|
| ============================================================================
|
"

STObject subclass: #WorkspaceController
		 instanceVariableNames: 'panel scrolledText evalButton resultButton'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

WorkspaceController comment: 'Controls a panel for entering smalltalk code and
evaluatin it' !

!WorkspaceController methodsFor: 'Initialization'!

init
	" Console will be a global varialbe where all error messages and
	  print, printNl will go to "
	| rect size contentView |

	rect _ Rect alloc initFromX: 500.0 y: 100.0 width: 400.0 height: 300.0.

	super init.
	self initPanel: rect.
	contentView _ panel contentView.
	contentView getBounds: (rect asCPointer).
	contentView notifyAncestorWhenFrameChanged: YES.
	contentView setAutoresizeSubviews: YES.
	rect setHeight: ((rect height) - 60.0).
	self initTextScrollView: rect.
	rect setY: (rect height); setHeight: 60.0.
	self initButton: rect.
	contentView addSubview: scrolledText id.
	contentView addSubview: evalButton id.
!

initPanel: rect
	panel _ Window alloc initContent: (rect asCPointer)
								style: 6
							  backing: 2
						   buttonMask: (4 bitOr: 1)
								defer: YES.
	panel setTitle: 'Workspace'.
!

initButton: rect
	evalButton _ Button alloc initFrame: (rect asCPointer).
	evalButton setTitle: 'Evaluate contents'.
	evalButton setImage: ((NXImage findImageNamed: 'Eval.tiff') id).
	evalButton setIconPosition: 2. 				"NX_ICONLEFT"
	evalButton setTarget: (self id).
	evalButton setAction: (OCInterface getUid: 'action1:').
	evalButton setAutosizing: (2 bitOr: 8).		" NX_WIDTHSIZABLE | NX_MAXYMARGINSIZABLE "
!

initTextScrollView: rect
	| font |
	scrolledText _ TextScrollView alloc initFrame: rect.
	scrolledText setAutosizing: ((8 bitOr: 16) bitOr: 2).	" NX_WIDTHSIZABLE | NX_HEIGHTSIZABLE | NX_MINYMARGINSIZABLE"
	font _ Font newFont: 'Ohlfs' size: 10.0.
	(scrolledText text) setMonoFont: YES.
	(scrolledText text) setFont: (font id).
	scrolledText setBorderType: 0.
!!


!WorkspaceController methodsFor: 'accessing'!

panel	
	^panel
!!

!WorkspaceController methodsFor: 'Target Action'!

action1: sender
	self evaluateContents.
!!

!WorkspaceController methodsFor: 'Evaluating Code'!

evaluateContents
	" Takes the contents of the workspaces text object and
	  compiles them into objects method executeStatements
	  and then executes it.  The result in shown is appended
	  to the text object. "
	| code len codeStream result text |
	text _ scrolledText text.
	len _ text textLength.
	code _ String new: (len + 1).  	
	text getSubstring: code start: 0 length: len.
	codeStream _ WriteStream on: (String new: 20).
	codeStream nextPutAll: 'executeStatements
	'; 
		nextPutAll: code.	
	Object class compile: (codeStream contents).
	codeStream emptyStream.
	result _ Object executeStatements.
	result printOn: codeStream.
	text appendText: '
-- Result --
'.
	text appendText: (codeStream contents).
!!

"
|
| ============================================================================
|
"

STObject subclass: #ConsoleController
		 instanceVariableNames: 'panel scrolledText'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

ConsoleController comment: 'Controls the display of out from interpreter'!

!ConsoleController methodsFor: 'Hacking'!

makeConsoleGlobal
	" This is used that object print can be redirected
	  to the workspace instead of going out stdout "
	Smalltalk at: #Console put: self
!

replaceObjectsPrint
	" This methods replaces Object print"

	Object compile:'
print
	| stream text |
	text _ Console text.
    stream _ WriteStream on: (String new: 0).
	self printOn: stream.
	text appendText: stream contents.
	text display.'.

	Object compile:'
error: message
	message printNl.'.

!!

!ConsoleController methodsFor: 'Initialization'!

init
	" Console will be a global varialbe where all error messages and
	  print, printNl will go to "
	| rect |
	rect _ Rect alloc initFromX: 400.0 y: 410.0 width: 400.0 height: 300.0.
	super init.
	self initPanel: rect.
	self initTextScrollView: rect.
	self compose.
	"
	self makeConsoleGlobal.
	self replaceObjectsPrint.
	"
!

initPanel: rect
	panel _ Window alloc initContent: (rect asCPointer)
								style: 6
							  backing: 2
						   buttonMask: (4 bitOr: 1)
								defer: YES.
	panel setTitle: 'Console'.
!

compose
	| contentView |
	contentView _ panel contentView.
	contentView notifyAncestorWhenFrameChanged: YES.
	contentView setAutoresizeSubviews: YES.
	contentView addSubview: scrolledText id.
!

initTextScrollView: rect
	scrolledText _ TextScrollView alloc initFrame: rect.
	scrolledText setBorderType: 0.
!!

!ConsoleController methodsFor: 'Accessing'!

text
	^(scrolledText text)
!

panel 
	^panel
!!

"
|
| ============================================================================
|
"

STObject subclass: #MethodBrowserController
		 instanceVariableNames: 'scrolledText browser splitView window browseTypeMatrix
		 						 methodBrowserDelegate buttonMatrix class box'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

MethodBrowserController comment: 'Created whenever a class is double clicked on'!

!MethodBrowserController methodsFor: 'Intialization'!

init
	| rect |
	super init.
	rect _ Rect alloc initFromX: 200.0 y: 200.0 width: 500.0 height: 400.0.
	self initWindow: rect.
	self initBrowser: rect.
	self initScrolledText: rect.
	self initSplitView: rect.
	self initButtonMatrix: rect.
	self initBrowseTypeButtonMatrix: rect.
	self compose.
!

initWithClass: aClass
	" This is the designated initializer for this Object "
	class _ aClass.
	self init.
!

initButtons: rect
	^self
!

initBrowser: rect
	browser _ NXBrowser alloc initFrame: (rect asCPointer).
	methodBrowserDelegate _ MethodBrowserDelegate alloc init.
	methodBrowserDelegate browseClass: class.
	browser setDelegate: (methodBrowserDelegate id).
	browser setTarget: (self id).
	browser setAction: (OCInterface getUid: 'action1:').
	browser setEmptySelectionEnabled: NO.
	browser setHorizontalScrollButtonsEnabled: NO.
	browser setHorizontalScrollerEnabled: NO.
	browser setMaxVisibleColumns: 2.
	browser loadColumnZero.
!


initSplitView: rect
	| bounds contentView size |
	bounds _ Rect alloc initFromRect: (rect asCPointer).	
	bounds setX: 0.0 ; setY: 0.0; setHeight: ((rect height) - 30.0).
	box _ Box alloc initFrame: (bounds asCPointer).
	box notifyAncestorWhenFrameChanged: YES.
	box setAutoresizeSubviews: YES.
	box setAutosizing: (2 bitOr:8; bitOr: 16).
	box setTitlePosition: 0.						" 0 = No Title "
	box setBorderType: 0.							" 0 = No Border "
	splitView _ NXSplitView alloc initFrame: (bounds asCPointer).
	box setContentView: (splitView id).
!

initWindow: rect
	window _ Window alloc initContent: (rect asCPointer)
								style: 6
							  backing: 2
						   buttonMask: (4 bitOr: 1)
								defer: YES.
	window setTitle: 'Method Browser for ', class name.
!

initScrolledText: rect
	| font |
	scrolledText _ TextScrollView alloc initFrame: rect.
	font _ Font newFont: 'Ohlfs' size: 10.0.
	(scrolledText text) setMonoFont: YES.
	(scrolledText text) setFont: (font id).
!

initBrowseTypeButtonMatrix: aRect
	| bounds size font |
	
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	size _ Size alloc init.
	bounds setX: 6.0 ; setY: ((aRect height) - 30.0) ; setHeight: 24.0 ; setWidth: ((aRect width) - 12.0).
	
	browseTypeMatrix _ Matrix alloc initFrame: (bounds asCPointer)
							mode: 2							"NX_LISTMODE"
							cellClass: (ButtonCell isa)
							numRows: 1
							numCols: 3.
	browseTypeMatrix setAutosizing: (2 bitOr: 8).
	browseTypeMatrix setAutosizeCells: YES.
	browseTypeMatrix setIntercell: (size asCPointer).
	
	font _ Font newFont: 'Helvetica-Bold' size: 10.0.
	browseTypeMatrix setFont: font id.
	
	browseTypeMatrix setAction: (OCInterface getUid: 'action2:').
	browseTypeMatrix setTarget: (self id).
		
	browseTypeMatrix setTitle: 'Browse Methods:' at: 0 anArg: 0.
	browseTypeMatrix setTitle: 'Instance' at: 0 anArg: 1.
	browseTypeMatrix setTitle: 'Class' at: 0 anArg: 2.
	"browseTypeMatrix setTitle: 'Both' at: 0 anArg: 3."
	
	(browseTypeMatrix cellAt: 0 anArg: 0) setEnabled: NO.
	(browseTypeMatrix cellAt: 0 anArg: 0) setBordered: NO.
	(browseTypeMatrix selectCellAt: 0 anArg: 1).
!

initButtonMatrix: aRect
	| bounds size font |
	
	bounds _ Rect alloc initFromRect: (aRect asCPointer).
	size _ Size alloc init.
	bounds setX: 0.0 ; setY: 0.0 ; setHeight: 60.0; setWidth: 180.0. "((bounds width) - 10.0)."
	
	buttonMatrix _ Matrix alloc initFrame: (bounds asCPointer)
							mode: 0							"NX_LISTMODE"
							cellClass: (ButtonCell isa)
							numRows: 1
							numCols: 6.
	buttonMatrix setAutosizeCells: YES.
	buttonMatrix setIntercell: (size asCPointer).
	
	font _ Font newFont: 'Helvetica-BoldOblique' size: 10.0.
	buttonMatrix setFont: font id.

	buttonMatrix setAction: (OCInterface getUid: 'action3:').
	buttonMatrix setTarget: (self id).
		
	buttonMatrix setTitle: 'Compile' at: 0 anArg: 0.
	buttonMatrix setTitle: 'Add' at: 0 anArg: 1.
	buttonMatrix setTitle: 'Remove' at: 0 anArg: 2.
	buttonMatrix setTitle: 'Add Cata...' at: 0 anArg: 3.
	buttonMatrix setTitle: 'File Out Cat...' at: 0 anArg: 4.
!

compose
	| contentView |
	" Arranges everything on screen so that it looks nice "
	
	contentView _ window contentView.
	contentView notifyAncestorWhenFrameChanged: YES.
	contentView setAutoresizeSubviews: YES.
	splitView addSubview: browser id.
	splitView addSubview: buttonMatrix id.
	splitView addSubview: scrolledText id.
	contentView addSubview: box id.
	contentView addSubview: browseTypeMatrix id.
	contentView display.
!

initButtons: rect
	^self
!!
	
!MethodBrowserController methodsFor: 'Action'!

action1: sender
	| cell selectedSel methodString text |

	text _ scrolledText text.
	cell _ browser selectedCell.
	selectedSel _ cell stringValue asSymbol.
	
	(class canUnderstand: selectedSel)
		ifTrue:  [
			methodString _ class sourceCodeAt: selectedSel.
			text setText: methodString.
			text sizeToFit.
			text scrollSelToVisible.
			scrolledText display.
			^self ].
			
	((class class) canUnderstand: selectedSel)
		ifTrue:  [
			methodString _ (class class) sourceCodeAt: selectedSel.
			text setText: methodString.
			text sizeToFit.
			text scrollSelToVisible.
			scrolledText display ].
!

action2: sender
	| matrix tag |
	sender _ Matrix newWithId: sender.
	tag _ sender selectedCol.
	
	tag == 1
		ifTrue: [ methodBrowserDelegate browseInstanceMethods. ].
	tag == 2
		ifTrue: [ methodBrowserDelegate browseClassMethods. ].
	
	browser loadColumnZero.
!

action3: sender
	| matrix tag |
	sender _ Matrix newWithId: sender.
	tag _ sender selectedCol.
	
	tag == 0
		ifTrue: [ self compileMethod. ]. 
	tag == 1
		ifTrue: [ self addMethod. ]. 
	tag == 4
		ifTrue: [ self fileOutCategory. ]. 
!!	

!MethodBrowserController methodsFor: 'Creating New Code'!

compileMethod
	| code len codeStream result text cell category |
	text _ scrolledText text.
	len _ text textLength.
	code _ String new: (len + 1).  	
	text getSubstring: code start: 0 length: len.
	
	cell _ ((browser matrixInColumn: 0) selectedCell).
	cell notNil
		ifTrue: [ class compile: code classified: (cell stringValue) ]
		ifFalse: [ class compile: code classified: 'New Category' ]
!

addMethod
	| text template |
	text _ scrolledText text.
	text setText: 'newMethod: anArg
	" A comment "
	^self'.
	scrolledText display.		
!

addCategory
!

fileOutCategory

!!

!MethodBrowserController methodsFor: 'Accessing'!

window
	^window

!!

"
|
| ============================================================================
|
"

STAppDelegate subclass: #AppDelegate
		 instanceVariableNames: 'mainMenu browserController app workspace console'
		 classVariableNames: ''
		 poolDictionaries: ''
		 category: nil
!

AppDelegate comment: 'Creates everything after the Application has been launched'!

!AppDelegate methodsFor: 'App Delegation'!

appDidInit: arg0
	app _ Application new.
	self initMainMenu.
	browserController _ ClassBrowserController alloc init.
	workspace _ WorkspaceController alloc init.
	console _ ConsoleController alloc init.
!!

!AppDelegate methodsFor: 'Initialising the Interface'!

initMainMenu
	| menuItem services windows info system |
	mainMenu _ Menu alloc initTitle: 'Smalltalk'.
	
	menuItem _ mainMenu addItem: 'Info' action: nil keyEquivalent: nil.
	mainMenu setSubmenu: ((self initInfoMenu) id) forItem: (menuItem id).
	menuItem _ mainMenu addItem: 'System' action: nil keyEquivalent: nil.
	mainMenu setSubmenu: ((self initSystemMenu) id) forItem: (menuItem id).
	menuItem _ mainMenu addItem: 'Edit' action: nil keyEquivalent: nil.
	mainMenu setSubmenu: ((self initEditMenu) id) forItem: (menuItem id).
	menuItem _ mainMenu addItem: 'Tools' action: nil keyEquivalent: nil.
	mainMenu setSubmenu: ((self initToolMenu) id) forItem: (menuItem id).

	"
	menuItem _ mainMenu addItem: 'Windows' action: nil keyEquivalent: nil.
	windows _ Menu alloc initTitle: 'Windows'.
	windows addItem: 'Arrange in Front' action: (OCInterface getUid: 'arrangeInFront:') keyEquivalent: nil.
	windows addItem: 'Miniaturize Window' action: (OCInterface getUid: 'miniaturize:') keyEquivalent: ($m asciiValue).
	windows addItem: 'Close Window' action: (OCInterface getUid: 'performClose:') keyEquivalent: ($W asciiValue).
	mainMenu setSubmenu: (windows id) forItem: (menuItem id).
	app setWindowsMenu: windows id.
	"
	menuItem _ mainMenu addItem: 'Services' action: nil keyEquivalent: nil.
	services _ Menu alloc initTitle: 'Services'.
	mainMenu setSubmenu: (services id) forItem: (menuItem id).
	app setServicesMenu: (services id).
	

	menuItem _ mainMenu addItem: 'Print...' action: nil keyEquivalent: ($p asciiValue).
	menuItem _ mainMenu addItem: 'Hide' action: (OCInterface getUid: 'hide:') keyEquivalent: ($h asciiValue).
	menuItem _ mainMenu addItem: 'Quit' action: (OCInterface getUid: 'terminate:') keyEquivalent: ($q asciiValue).	
	
	app setMainMenu: (mainMenu id).
!
			
initInfoMenu
	| infoMenu |
	infoMenu _ Menu alloc initTitle: 'Info'.
	infoMenu addItem: 'Info...' action: nil keyEquivalent: nil.
	infoMenu addItem: 'Preferences...' action: nil keyEquivalent: nil.
	infoMenu addItem: 'Help...' action: (OCInterface getUid: 'showHelpPanel:') keyEquivalent: ($? asciiValue).
	^infoMenu	
!
		
initSystemMenu
	| systemMenu menuItem |
	systemMenu _ Menu alloc initTitle: 'System'.
	systemMenu addItem: 'Snapshot...' action: (OCInterface getUid: 'action3:') keyEquivalent: ($s asciiValue).
	systemMenu addItem: 'File In...' action: (OCInterface getUid: 'action4:') keyEquivalent: ($i asciiValue).
	systemMenu addItem: 'File Out...' action: (OCInterface getUid: 'action5:') keyEquivalent: ($o asciiValue).
	^systemMenu	
!
		
initEditMenu
	| editMenu |
	editMenu _ Menu alloc initTitle: 'Edit'.
	" Don't give them menuitems a specific target that way they will
	  be forwarded to first responder "
	editMenu addItem: 'Cut' action: (OCInterface getUid: 'cut:') keyEquivalent: ($x asciiValue).
	editMenu addItem: 'Copy' action: (OCInterface getUid: 'copy:') keyEquivalent: ($c asciiValue).
	editMenu addItem: 'Paste' action: (OCInterface getUid: 'paste:') keyEquivalent: ($v asciiValue).
	^editMenu	
!
		
initToolMenu
	| toolMenu menuItem |
	toolMenu _ Menu alloc initTitle: 'Edit'.
	toolMenu addItem: 'Workspace...' action: (OCInterface getUid: 'action1:') keyEquivalent: ($w asciiValue).
	toolMenu addItem: 'Console...' action: (OCInterface getUid: 'action2:') keyEquivalent: ($C asciiValue).
	toolMenu addItem: 'File Browser...' action: nil keyEquivalent: ($f asciiValue).
	^toolMenu		
!

action1: sender
	" This is the action called when someone wants to open the
	  workspace panel "
	| panel |
	panel _ workspace panel.
	panel makeKeyAndOrderFront: (panel id).
	panel display.
!

action2: sender
	" This is the action called when someone wants to open the
	  workspace panel "
	| panel |
	panel _ console panel.
	panel makeKeyAndOrderFront: (panel id).
	panel display.
!

action3: sender
	" Called by menu item to save the image "
	Smalltalk snapshot.
!

action4: sender
	" Action for filing in "
!

action5: sender
	" Action for filing out "
	| savePanel value file dir class rect |

	rect _ Rect alloc initFromX: 200.0 y: 200.0 width: 500.0 height: 400.0.

	" For some reason #new won't work "
	savePanel _ SavePanel newContent: (rect asCPointer)
							   style: 6
							 backing: 2
						  buttonMask: (4 bitOr: 1)
							   defer: YES.
							   
	savePanel setRequiredFileType: 'st'.
	savePanel setTitle: 'Filing Out'.
	value _ savePanel runModal. "ForDirectory: '/' file: 'AClass.st' ."

	value == 1
		ifTrue: [
			file _ savePanel filename. 
			dir _ savePanel directory.
			dir, '/', file printNl. ].

!!


"
|
| ============================================================================
|
"

!Object class methodsFor: 'main'!

main
	| app appDelegate |
	app _ Application new.
	appDelegate _ AppDelegate alloc init.
	app setDelegate: (appDelegate id).
	app setAutoupdate: YES.
	app run.
!!

Object main !