"======================================================================
|
|   X Colormap class.
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     19 Oct 90	  created.
|
"

XServerObject subclass: #Colormap
	  instanceVariableNames: ''
	  classVariableNames: ''
	  poolDictionaries: ''
	  category: 'X Objects'
!

!Colormap class methodsFor: 'X object instance creation'!

copyColormapAndFree: srcCmap
    | id req |
    id _ srcCmap server nextId.
    req _ XRequest fixedCommand: 80.
    req ulong: id; ulong: srcCmap id.
    req queueOn: srcCmap server.
    ^self new init: id onServer: srcCmap server
!

createColormap: visual window: window alloc: alloc
    | id req |
    id _ window server nextId.
    req _ XRequest fixedCommand: 78 
		   aux: (XProtocol map: alloc into: #(None All)).
    req ulong: id; ulong: window id; ulong: visual id.
    req queueOn: window server.
    ^self new init: id onServer: window server
! !


!Colormap methodsFor: 'X protocol requests'!

allocColor: red green: green blue: blue
    | req redVar greenVar blueVar pixelVar |
    req _ XRequest fixedCommand: 84.
    req long: id; uword: red; uword: green; uword: blue.
    req queueOn: server.
    
    server getReply.
    redVar _ server uword.
    greenVar _ server uword.
    blueVar _ server uword.
    server uword.		"ignored"
    pixelVar _ server ulong.
    server skip: 12.
    ^AllocColorReply red: redVar green: greenVar blue: blueVar pixel: pixelVar
!

allocColorCells: colorsVar planes: planesVar contiguous: contiguousVar
    | req numPixels numMasks pixels masks |
    req _ XRequest fixedCommand: 86 aux: contiguousVar.
    req long: id; uword: colorsVar; uword: planesVar.
    req queueOn: server.
    
    server getReply.
    numPixels _ server uword.
    numMasks _ server uword.
    server skip: 20.
    pixels _ Array new: numPixels.
    masks _ Array new: numMasks.
    1 to: numPixels do:
	[ :i | pixels at: i put: server ulong ].
    1 to: numMasks do:
	[ :i | masks at: i put: server ulong ].
    ^AllocColorCellsReply pixels: pixels masks: masks
!    

allocColorPlanes: colorsVar reds: redsVar greens: greensVar blues: bluesVar
    contiguous: contiguousVar 
    | req numPixels pixels masks redMask greenMask blueMask |
    req _ XRequest fixedCommand: 87 aux: contiguousVar.
    req long: id; uword: colorsVar; uword: redsVar; uword: greensVar;
	uword: bluesVar.
    req queueOn: server.
    
    server getReply.
    numPixels _ server uword.
    server uword.		"unused"
    redMask _ server ulong.
    greenMask _ server ulong.
    blueMask _ server ulong.
    server skip: 8.
    pixels _ Array new: numPixels.
    1 to: numPixels do:
	[ :i | pixels at: i put: server ulong ].
    ^AllocColorPlanesReply redMask: redMask greenMask: greenMask 
			   blueMask: blueMask pixels: pixels
!    

allocNamedColor: nameVar
    | req pixel exactRed exactGreen exactBlue visualRed visualGreen 
      visualBlue |
    req _ XRequest command: 85.
    req long: id; uword: nameVar size; uword: 0; string: nameVar.
    req queueOn: server.
    
    server getReply.
    pixel _ server ulong.
    exactRed _ server uword.
    exactGreen _ server uword.
    exactBlue _ server uword.
    visualRed _ server uword.
    visualGreen _ server uword.
    visualBlue _ server uword.
    server skip: 8.
    ^AllocNamedColorReply pixel: pixel exactRed: exactRed 
			  exactGreen: exactGreen exactBlue: exactBlue 
			  visualRed: visualRed visualGreen: visualGreen 
			  visualBlue: visualBlue
!    

freeColormap
    | req |
    req _ XRequest fixedCommand: 79.
    req ulong: id.
    req queueOn: server
!

freeColors: pixels planeMask: planeMask
    | req |
    req _ XRequest command: 88.
    req ulong: id; ulong: planeMask.
    pixels do: [ :pixel | req ulong: pixel ].
    req queueOn: server
!

installColormap
    | req |
    req _ XRequest fixedCommand: 81.
    req ulong: id.
    req queueOn: server
!

lookupColor: name
    | req exactRed exactGreen exactBlue visualRed visualGreen visualBlue |
    req _ XRequest command: 92.
    req long: id; uword: name size; uword: 0; string: name.
    req queueOn: server.
    
    server getReply.
    exactRed _ server uword.
    exactGreen _ server uword.
    exactBlue _ server uword.
    visualRed _ server uword.
    visualGreen _ server uword.
    visualBlue _ server uword.
    server skip: 12.
    ^LookupColorReply exactRed: exactRed exactGreen: exactGreen
		      exactBlue: exactBlue visualRed: visualRed
		      visualGreen: visualGreen visualBlue: visualBlue
!

queryColors: pixels
    | req colors numRgbs |
    req _ XRequest command: 91.
    req long: id.
    pixels do: [ :pixel | req ulong: pixel ].
    req queueOn: server.
    
    server getReply.
    numRgbs _ server word.    
    server skip: 22.
    colors _ Array new: numRgbs.
    1 to: numRgbs do: [ :i | colors at: i put: server rgb ].
    ^QueryColorsReply colors: colors
!

storeColors: items
    | req |
    req _ XRequest command: 89.
    req ulong: id.
    items do: [ :item | req colorItem: item ].
    req queueOn: server.
    
!

storeNamedColor: pixel name: name doRed: doRed doGreen: doGreen doBlue: doBlue
    | req dos |
    dos _ 0.
    doRed ifTrue: [ dos _ dos bitOr: 16r01].
    doGreen ifTrue: [ dos _ dos bitOr: 16r02].
    doBlue ifTrue: [ dos _ dos bitOr: 16r04].
    req _ XRequest command: 90 aux: dos.
    req ulong: id; ulong: pixel; word: name size; word: 0; string: name.
    req queueOn: server
!

uninstallColormap
    self justId: 82
! !

LoadStatus _ 'Colormap'!

