"======================================================================
|
|   X Drawable classes
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     26 Oct 90	  Began restructuring.
|
| sbyrne     24 May 90	  created.
|
"

XServerObject subclass: #Drawable
	  instanceVariableNames: ''
	  classVariableNames: ''
	  poolDictionaries: ''
	  category: 'X Objects'
!


!Drawable methodsFor: 'X protocol requests'!

copyArea: srcDrawable gc: gc srcX: srcX srcY: srcY width: width height: height
    dstX: dstX dstY: dstY
    | req |
    req _ XRequest fixedCommand: 62.
    req ulong: srcDrawable id; ulong: id; ulong: gc id; word: srcX; word: srcY;
	word: dstX; word: dstY; uword: width; uword: height.
    req queueOn: server
!

copyPlane: srcDrawable gc: gc srcX: srcX srcY: srcY width: width 
    height: height dstX: dstX dstY: dstY bitPlane: bitPlane
    | req |
    req _ XRequest fixedCommand: 63.
    req ulong: srcDrawable id; ulong: id; ulong: gc id; word: srcX; word: srcY;
	word: dstX; word: dstY; uword: width; uword: height; ulong: bitPlane.
    req queueOn: server
!

fillPoly: gc shape: shape coordinateMode: coordMode
    points: points
    | req |
    req _ XRequest command: 69.
    req ulong: id; ulong: gc id; 
	byte: (XProtocol map: shape into: #(Complex Nonconvex Convex));
	byte: (XProtocol map: coordMode into: #(Origin Previous));
	word: 0.
    points do:
	[ :point | req point: point ].
    req queueOn: server
!

getGeometry
    | req root depth x y width height borderWidth |
    req _ XRequest fixedCommand: 14.
    req ulong: id.
    req queueOn: server.

    depth _ server getReply.
    root _ Window idFrom: server.
    x _ server word.
    y _ server word.
    width _ server uword.
    height _ server uword.
    borderWidth _ server uword.    
    server skip: 10.
    ^GetGeometryReply depth: depth root: root x: x y: y width: width
		      height: height borderWidth: borderWidth
!

getImage: x y: y width: width height: height planeMask: planeMask 
    format: format
    | req depth visual data len |
    req _ XRequest fixedCommand: 73
		   aux: (XProtocol map: format into: #(nil XYPixmap ZPixmap)).
    req ulong: id; word: x; word: y; uword: width; uword: height;
	ulong: planeMask.
    req queueOn: server.

    server getReplyAndLength: 
	[ :lenLongs :depthByte | len _ lenLongs * 4.
				 depth _ depthByte ].
    visual _ Visual maybeIdFrom: server symbols: #(None).
    server skip: 20.
    data _ server getBytes: len.
    " ### the bytes needs to be more precisely determined"
    ^GetImageReply depth: depth visual: visual data: data
!

imageText8: gc x: x y: y string: string
    | req |
    req _ XRequest command: 76 aux: string size.
    req ulong: id; ulong: gc id; word: x; word: y; string: string.
    req queueOn: server
!

imageText16: gc x: x y: y string: string
    | req |
    req _ XRequest command: 77 aux: string size.
    req ulong: id; ulong: gc id; word: x; word: y; string16: string.
    req queueOn: server
!

polyArc: gc arcs: arcs
    | req |
    req _ XRequest command: 68.
    req ulong: id; ulong: gc id.
    arcs do: [ :arc | req arc: arc ].
    req queueOn: server
!

polyFillArc: gc arcs: arcs
    | req |
    req _ XRequest command: 71.
    req ulong: id; ulong: gc id.
    arcs do: [ :arc | req arc: arc ].
    req queueOn: server
!

polyFillRectangle: gc rectangles: rectangles
    | req |
    req _ XRequest command: 70.
    req ulong: id; ulong: gc id.
    rectangles do: [ :rect | req rectangle: rect ].
    req queueOn: server
!

polyLine: gc coordinateMode: coordMode points: points
    | req |
    req _ XRequest command: 65
		   aux: (XProtocol map: coordMode into: #(Origin Previous)).
    req ulong: id; ulong: gc id.
    points do: [ :point | req point: point ].
    req queueOn: server
!

polyPoint: gc coordinateMode: coordMode points: points
    | req |
    req _ XRequest command: 64
		   aux: (XProtocol map: coordMode into: #(Origin Previous)).
    req ulong: id; ulong: gc id.
    points do: [ :point | req point: point ].
    req queueOn: server
!

polyRectangle: gc rectangles: rectangles
    | req |
    req _ XRequest command: 67.
    req ulong: id; ulong: gc id.
    rectangles do: [ :rect | req rectangle: rect ].
    req queueOn: server
!

polySegment: gc segments: segments
    | req |
    req _ XRequest command: 66.
    req ulong: id; ulong: gc id.
    segments do: [ :seg | req segment: seg ].
    req queueOn: server
!

polyText8: gc x: x y: y items: items
    | req |
    req _ XRequest command: 74.
    req ulong: id; ulong: gc id; word: x; word: y.
    items do: [ :item | req textItem8: item ].
    req queueOn: server
!

polyText16: gc x: x y: y items: items
    | req |
    req _ XRequest command: 75.
    req ulong: id; ulong: gc id; word: x; word: y.
    items do: [ :item | req textItem16: item ].
    req queueOn: server
!

putImage: gc depth: depth width: width height: height dstX: dstX dstY: dstY
    leftPad: leftPad format: format data: data
    | req |
    req _ XRequest command: 72
		   aux: (XProtocol map: format 
				   into: #(Bitmap XYPixmap ZPixmap)).
    req ulong: id; ulong: gc id; uword: width; uword: height; word: dstX;
	word: dstY; ubyte: leftPad; ubyte: depth; word: 0;
	bytes: data.
    req queueOn: server
!


queryBestSize: class width: width height: height
    | req optWidth optHeight |
    req _ XRequest fixedCommand: 97
		   aux: (XProtocol map: class into: #(Cursor Tile Stipple)).
    req ulong: id; uword: width; uword: height.
    req queueOn: server.
    
    server getReply.
    optWidth _ server uword.
    optHeight _ server uword.
    server skip: 20.
    ^QueryBestSizeReply width: optWidth height: optHeight
!!


LoadStatus _ 'Drawable'!


FileStream fileIn: 'Pixmap.st'!
FileStream fileIn: 'Window.st'!

