"======================================================================
|
|   A window containing a list of items, part of the STIX based GUI 
|   building blocks. 
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     20 Sep 94	  Fixed per Joerg Rade to remove onServer: from
|			  initWin: pixmap creation code.
|
| sbb	     16 Feb 92	  created around christmas 91
|
"
FileStream fileIn: 'EventHandler.st'!
FileStream fileIn: 'Widget.st'!

Widget subclass: #ListWin
       instanceVariableNames: 'background array gc xorGC
                               font fontInfo 
                               height winWidth winHeight
                               curIndex' "I'm not sure that we need curIndex"

       classVariableNames: ''
       poolDictionaries: ''
       category: 'Window base class'
!

!ListWin class methodsFor: 'Instance creation'!

on: array server: server 
    | rootWin |
    rootWin _ (server root screens at: 1)  window.    
    ^self newWindowUsing: array onServer: server parent: rootWin.
!

on: array server: server parent: parent
    ^self newWindowUsing: array onServer: server parent: parent
!

newWindowUsing: array onServer: server parent: parent 
    | w winWidth winHeight font fontInfo height |
    font _ self defaultFontFor: server.
    fontInfo _ font queryFont.
    height _ fontInfo fontAscent + fontInfo fontDescent.
    winHeight _ height * array size.
    winWidth _ self maxWidthFor: array inFont: fontInfo.
    w _ self newWindowOn: server parent: parent
	     width: winWidth height: winHeight.
    w init: array height: height
      winWidth: winWidth winHeight: winHeight
      font: font.
    ^w
!

newWindowUsing: array onServer: server at: aPoint parent: parent 
    | w winWidth winHeight font fontInfo height |
    font _ self defaultFontFor: server.
    fontInfo _ font queryFont.
    height _ fontInfo fontAscent + fontInfo fontDescent.
    winHeight _ height * array size.
    winWidth _ self maxWidthFor: array inFont: fontInfo.
    w _ self newWindowOn: server parent: parent
	     x: aPoint x y: aPoint y
	     width: winWidth height: winHeight.
    w init: array height: height
      winWidth: winWidth winHeight: winHeight
      font: font.
    ^w
!

newWindowOn: server parent: parent inRect: aRect
    ^self newWindowOn: server parent: parent 
	  x: aRect left y: aRect top
	  width: aRect width height: aRect height
!

newWindowOn: server parent: parent width: winWidth height: winHeight
    ^self newWindowOn: server parent: parent
	  x: 0 y: 0
	  width: winWidth height: winHeight
!

newWindowOn: server parent: parent x: x y: y width: winWidth height: winHeight
    | screen |
    screen _ server root screens at: 1.
    ^self createWindow: parent
	  class: #InputOutput
	  depth: 0		"default to depths of root"
	  visual: #CopyFromParent 
	  x: x y: y
	  width: winWidth height: winHeight 
	  borderWidth: self defBorderWidth
	  onServer: server
	  attrs: [ :r | self winAttributesFor: r onScreen: screen ]
!

defaultFontFor: server
    "### This should cache its result, and should really come from the server"
    ^Font openFont: '9x15' onServer: server.
!

defBorderWidth
    ^5
!

winAttributesFor: r onScreen: screen
    r eventMask: self eventSet;
	backgroundPixel: screen blackPixel
!

eventSet
    ^#(Exposure ButtonPress 
	   ButtonRelease
	   Button1Motion
	   "StructureNotify"
	   )
!


maxWidthFor: array inFont: fontInfo
    | width |
    width _ 0.
    array do: [ :string | width _ width max: (self measure: string 
						   withFont: fontInfo) ].
    ^width
!

"This should probably be a method on a font"

measure: aString withFont: fontInfo
    | charInfo totalWidth ch min charInfos |
    "Need to check minByte & maxByte properly"
    (fontInfo minByte1 ~= 0
	or: [ fontInfo maxByte1 ~= 0 ])
	ifTrue: [ ^self error: 'unhandled case: minByte or maxByte are nonzero' ].
    totalWidth _ 0.
    min _ fontInfo minCharOrByte2.
    charInfos _ fontInfo charInfos.
    aString do:
	[ :char | ch _ char asciiValue - min.
		  charInfo _ charInfos at: ch + 1.
		  totalWidth _ totalWidth + charInfo characterWidth ].
    ^totalWidth

! !


!ListWin methodsFor: 'accessing'!

server
    ^server
!

fontInfo
    ^fontInfo
! !




!ListWin methodsFor: 'private'!

init: anArray height: aHeight winWidth: aWinWidth winHeight: aWinHeight 
    font: aFont
    | screen |
    array _ anArray.
    winWidth _ aWinWidth.
    winHeight _ aWinHeight.
    font _ aFont.
    height _ aHeight.
    fontInfo _ font queryFont.
    gc _ GC createGC: self onServer: server 
	    args: [ :r | r font: font;
			     graphicsExposures: false ].
    self initWin: array.
    screen _ server root screens at: 1.
    xorGC _ GC createGC: self onServer: server
	     args: [ :r | r function: #CopyInverted";
			      foreground: screen blackPixel;
			      background: screen whitePixel" ].
!

blitIt
    self copyArea: background gc: gc 
	 srcX: 0 srcY: 0
	 width: winWidth height: winHeight
	 dstX: 0 dstY: 0.
!

initWin: lineArray
    | y bgc screen |
    screen _ server root screens at: 1.
    array _ lineArray.
    background _ Pixmap createPixmap: self depth: 8
			width: winWidth height: winHeight.
    bgc _ GC createGC: background onServer: server
	     args: [ :r | r foreground: screen blackPixel;
			      graphicsExposures: false ].
    background polyFillRectangle: bgc
	       rectangles: (Array with: (0@0 corner: winWidth@winHeight)).
    y _ fontInfo fontAscent.
    array do: 
	[ :elt | background imageText8: gc
		     x: 0 y: y string: elt.
		 y _ y + height ].
    self setupHandlers.
!



trackPointer: e
    | index | 
    ((e eventY between: 0 and: winHeight)
	 and: [ (e eventX between: 0 and: winWidth) ])    
	ifTrue: [ index _ e eventY // height.
		  index ~= curIndex 
		      ifTrue: [ self unhighlight.
				curIndex _ index.
				self highlight. ].
		  ]
	ifFalse: [ self unhighlight.
		   curIndex _  nil ].

!


highlight
    self setHighlight: true.
!

unhighlight
    curIndex notNil
	ifTrue: [ self setHighlight: false ].
!

setHighlight: desiredValue
    | rect |
    true "(curIndex between: 0 and: maxY) "
	ifTrue: [ rect _ 0@(curIndex * height) extent: winWidth@height.
		  self polyFillRectangle: xorGC
		       rectangles: (Array with: rect). ].
!


motionNotify: e
    self trackPointer: e
!

pressed: e
    | button |
    button _ e detail.
    button == 1 ifTrue: [ self trackPointer: e ].
    button == 2 ifTrue: [ self testMsg: e ].
    button == 3 ifTrue: [ ExitHandlerLoop _ true].
!

released: e
    self unhighlight.
    curIndex notNil
	ifTrue: [ curIndex printNl. ].
    curIndex _ nil.
!

exposed: e
    e count = 0 ifTrue: [ self blitIt ] 
!

setupHandlers
    | w h |
    self registerFor: ExposeEvent
	when: [ :e | e window == id ]     do: [ :e | self exposed: e ].
    self registerFor: ButtonPressEvent
	when: [ :e | e event == id ]      do: [ :e | self pressed: e ].
    self registerFor: ButtonReleaseEvent
	when: [ :e | e event == id ]      do: [ :e | self released: e ].
    self registerFor: MotionNotifyEvent
	when: [ :e | e event == id ]      do: [ :e | self motionNotify: e ].
!

close
    self destroyWindow
!

testMsg: event
    | l |
    l _ array size.
    200 timesRepeat: 
	[ 
	  0 to: l do: 
	      [ :i | self unhighlight.
		     curIndex _ i.
		     self highlight ].
	  l to: 0 by: -1  do: 
	      [ :i | self unhighlight.
		     curIndex _ i.
		     self highlight ].
	  ].
!

testMsg: event
    | w rootWin |
    rootWin _ (server root screens at: 1)  window.    
    w _ ListWin newWindowUsing: #(Foo on you) onServer: server 
		at: (event rootX @ event rootY) parent: rootWin.
    w mapWindow
! !

