"======================================================================
|
|   Experimental STIX based GUI slider widget definition.
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     16 Feb 92	  created a while ago.
|
"
FileStream fileIn: 'Button.st'!

Widget subclass: #Slider
       instanceVariableNames: 'gc max min current parent upArrow downArrow
                               height thumbY thumbGc thumbBlack'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'Browsing experimental'
!

!Slider class methodsFor: 'Instance creation'!

on: win notifying: notifying max: max min: min initial: value
    | inst server screen xOrigin height ownerGeom pixmap | 
    server _ win server.
    screen _ server root screens at: 1.
    ownerGeom _ win getGeometry.
    xOrigin _ ownerGeom width - 20.
    height _ ownerGeom height.
    "self is broken here"
    pixmap _ self stripePix: server.
    inst _ self createWindow: win class: #InputOutput
		depth: 0
		visual: #CopyFromParent x: xOrigin y: 0
		width: 20 height: height borderWidth: 0 onServer: server
		attrs: [ :r | r eventMask: #(Exposure ButtonPress);
				  "winGravity: #SouthEast;
				  bitGravity: #Center;"
				  "backgroundPixel: screen whitePixel;"
				  backgroundPixmap: pixmap ].
    ^inst init: win notifying: notifying max: max min: min initial: value
		height: height - (20 * 2)
! !


!Slider methodsFor: 'notification'!

resized
    | ownerGeom |
    ownerGeom _ parent getGeometry.
    self configureWindow: [ :r | r height: ownerGeom height;
				     x: ownerGeom width - 20 ].
    downArrow configureWindow: [ :r | r y: ownerGeom height - 20 ].
    height _ ownerGeom height - (20 * 2).
    self drawThumb.
! !



!Slider methodsFor: 'private'!

init: parentWin notifying: notifying max: maxVal min: minVal initial: value
    height: anInteger
    max _ maxVal.
    min _ minVal.
    current _ value.
    notify _ notifying.
    parent _ parentWin.
    height _ anInteger.
    self initWin.
!

initWin
    | ownerGeom xOrigin height screen root rootWin w |
    screen _ server root screens at: 1.
    ownerGeom _ self getGeometry.
    self mapWindow.
    self createUpArrow: screen geom: ownerGeom.
    self createDownArrow: screen geom: ownerGeom.
    thumbGc _ GC createGC: self onServer: server
		 args: [ :r | r foreground: screen whitePixel ].
    thumbBlack _ GC createGC: self onServer: server
		 args: [ :r | r foreground: screen blackPixel ].
    self setupHandlers.
!

createUpArrow: screen geom: ownerGeom
    | height bgc points |
    height _ ownerGeom height.
    points _ OrderedCollection new add: 9@2;
	add: 16@17;
	add: 2@17;
	add: 9@2; yourself.
    bgc _ GC createGC: self onServer: server
	     args: [ :r | r foreground: screen blackPixel ].
    upArrow _ Button createOn: self in: (0@0 extent: 20@20)
		     notifying: self with: #upArrow: drawWith: 
			 [ :win | win polyLine: bgc coordinateMode: #Origin
				  points: points. ].
!

createDownArrow: screen geom: ownerGeom
    | height bgc points |
    height _ ownerGeom height.
    points _ OrderedCollection new add: 9@17;
	add: 16@2;
	add: 2@2;
	add: 9@17; yourself.
    bgc _ GC createGC: self onServer: server
	     args: [ :r | r foreground: screen blackPixel ].
    downArrow _ Button createOn: self in: (0@(height - 20) extent: 20@20)
		       notifying: self with: #downArrow: drawWith: 
			   [ :win | win polyLine: bgc coordinateMode: #Origin
				    points: points. ].
! 


upArrow: sender
    self setVal: current - 1.
    ^notify upArrow.
!

downArrow: sender
    self setVal: current + 1.
    ^notify downArrow.
!

setVal: aValue
    current _ (aValue max: min) min: max.
    self drawThumb.
!


drawThumb
    | delta yPos rect arc innerArc circle |
    delta _ max - min.		"may be + 1"
    yPos _ ((current asFloat / delta) * (height - 20)) truncated + 20.
    rect _ 0@yPos extent: 20@20.
    circle _ 360*64.
    arc _ Arc x: 1 y: yPos
	      width: 18 height: 18 angle1: 0 angle2: circle.
    innerArc _ Arc x: 2 y: yPos + 1
	      width: 17 height: 17 angle1: 0 angle2: circle.
    thumbY notNil
	ifTrue: [ self clearArea: 0 y: thumbY width: 20 height: 20 
		       exposures: false. ].
    self polyFillArc: thumbGc arcs: (Array with: innerArc).
    self polyArc: thumbBlack arcs: (Array with: arc).
    "self polyFillRectangle: thumbGc rectangles: (Array with: rect)."
    thumbY _ yPos.
!



setupHandlers
    self registerFor: ExposeEvent
	 when: [ :e | e window == id ]
	 do: [ :e | e count = 0
		       ifTrue: [ self drawThumb ] ].
!


initArrowHandlers: win
    win registerFor: ButtonPressEvent
      when: [ :e | e event == win id ]
      do: [ :e | e detail == 3
		     ifTrue: [ ExitHandlerLoop _ true ].
		 e detail == 1
		     ifTrue: [ self highLight: win ].
		     ].
    win registerFor: NoExposureEvent
	when: [ :e | true ]
	do: [ :e | nil ].
!


highLight: win
    'highlighting ... ' printNl.

! !
