"======================================================================
|
|   X window class
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     26 Oct 90	  Began rehacking.
|
| sbyrne     24 May 90	  created.
|
"

Drawable subclass: #Window
	 instanceVariableNames: ''
	 classVariableNames: ''
	 poolDictionaries: ''
	 category: 'X Objects'
!
    
Window comment: '###No comment yet'!



!Window class methodsFor: 'X object instance creation'!

createWindow: parent class: class depth: depth visual: visual x: x y: y
    width: width height: height borderWidth: borderWidth onServer: server
    attrs: aBlock
    | id req |
    id _ server nextId.
    req _ XWindowRequest command: 1 aux: depth.
    req ulong: id; ulong: parent id; 
	word: x; word: y;
	uword: width; uword: height; uword: borderWidth;
	word: (XProtocol map: class
		 into: #(CopyFromParent InputOutput InputOnly));
	long: (XProtocol maybeMapId: visual into: #(CopyFromParent)).

    aBlock notNil
	ifTrue: [ aBlock value: req ]
	ifFalse: [ req noBits ].
    req queueOn: server.
    ^self new init: id onServer: server
! !


!Window methodsFor: 'X protocol requests'!

changeProperty: property type: type format: format mode: mode data: data
    | req |
    req _ XRequest command: 18
		   aux: (XProtocol map: mode into: #(Replace Prepend Append)).
    req ulong: id; ulong: property id; ulong: type id; ubyte: format.
    req pad; ulong: (data size // (format // 8)).
    format == 8 ifTrue: 
	[ data do: [ :b | req ubyte: b ] ].
    format == 16 ifTrue: 
	[ data do: [ :w | req uword: w ] ] .
    format == 32 ifTrue: 
	[ data do: [ :l | req ulong: l ] ] .
    req queueOn: server
!

changeSaveSet: mode
    | req |
    req _ XRequest fixedCommand: 6
		   aux: (XProtocol map: mode into: #(Insert Delete)).
    req ulong: id.
    req queueOn: server
!

changeWindowAttributes: aBlock
    | req |
    req _ XWindowRequest command: 2.
    req ulong: id.
    aBlock notNil
	ifTrue: [ aBlock value: req ]
	ifFalse: [ req noBits ].
    req queueOn: server
!

circulateWindow: direction
    | req |
    req _ XRequest fixedCommand: 13
		   aux: (XProtocol map: direction into: #(RaiseLowest 
							     LowerHighest)).
    req ulong: id.
    req queueOn: server
!

clearArea: x y: y width: width height: height exposures: exposures
    | req |
    req _ XRequest fixedCommand: 61
		   aux: (XProtocol mapBool: exposures).
    req ulong: id; word: x; word: y; word: width; word: height.
    req queueOn: server
!

configureWindow: aBlock
    | req |
    req _ XConfigureWindowRequest command: 12.
    req ulong: id.
    aBlock notNil
	ifTrue: [ aBlock value: req ]
	ifFalse: [ req noBits ].
    req queueOn: server
!

deleteProperty: property
    | req |
    req _ XRequest fixedCommand: 19.
    req ulong: id; ulong: property id.
    req queueOn: server
!


destroySubwindows
    | req |
    req _ XRequest fixedCommand: 5.
    req ulong: id.
    req queueOn: server
!

destroyWindow
    | req |
    req _ XRequest fixedCommand: 4.
    req ulong: id.
    req queueOn: server
!

getMotionEvents: start stop: stop
    | req len events |
    req _ XRequest fixedCommand: 39.
    req ulong: id;
	ulong: (XProtocol maybeMap: start into: #(CurrentTime));
	ulong: (XProtocol maybeMap: stop  into: #(CurrentTime)).
    req queueOn: server.
    
    server getReply.
    len _ server ulong.
    server skip: 20.
    events _ Array new: len.
    1 to: len do: [ :i | events at: i
				put: server timeCoord ].
    ^GetMotionEvents events: events
!

getProperty: property type: type longOffset: longOffset longLength: longLength
    delete: delete
    | req format bytesAfter typeRep valLen value |
    req _ XRequest fixedCommand: 20 aux: (XProtocol mapBool: delete).
    req ulong: id; ulong: property id;
	ulong: (XProtocol maybeMapId: type into: #(AnyPropertyType));
	ulong: longOffset; ulong: longLength.
    req queueOn: server.

    format _ server getReply.
    typeRep _ Atom maybeIdFrom: server symbols: #(None).
    bytesAfter _ server ulong.
    valLen _ server ulong.
    server skip: 12.
    format = 0 ifTrue:
	[ value _ ByteArray new: 0 ].
    format = 8 ifTrue:
	[ value _ server getBytes: valLen ].
    format = 16 ifTrue:
	[ value _ Array new: valLen.
	  1 to: valLen do: [ :i | value at: i put: (server uword) ] ].
    format = 32 ifTrue:
	[ value _ Array new: valLen.
	  1 to: valLen do: [ :i | value at: i put: (server ulong) ] ].
    server skipPad.
    ^GetPropertyReply format: format type: typeRep bytesAfter: bytesAfter
		      value: value
!

getWindowAttributes
    | req backingStore visual class bitGravity winGravity backingPlanes
      backingPixel saveUnder mapIsInstalled mapState overrideRedirect colormap 
      allEventMasks yourEventMask doNotPropagateMask |
    req _ XRequest fixedCommand: 3.
    req ulong: id.
    req queueOn: server.
    
    backingStore _ XProtocol unmap: (server getReply)
			     into: #(NotUseful WhenMapped Always).
    visual _ Visual idFrom: server.
    class _ XProtocol unmap: (server uword) into: #(InputOutput InputOnly).
    bitGravity _ XProtocol unmap: (server ubyte) into: BitGravityNames.
    winGravity _ XProtocol unmap: (server ubyte) into: WinGravityNames.
    backingPlanes _ server ulong.
    backingPixel _ server ulong.
    saveUnder _ server bool.
    mapIsInstalled _ server bool.
    mapState _ XProtocol unmap: (server ubyte) into: #(Unmapped Unviewable 
							Viewable).
    overrideRedirect _ server bool.
    colormap _ Colormap maybeIdFrom: server symbols: #(None).
    allEventMasks _ server ulong.
    yourEventMask _ server ulong.
    doNotPropagateMask _ server uword.
    server skip: 2.
    ^GetWindowAttributesReply 
	backingStore: backingStore visual: visual winClass: class 
	bitGravity: bitGravity winGravity: winGravity
	backingPlanes: backingPlanes backingPixel: backingPixel 
	saveUnder: saveUnder mapIsInstalled: mapIsInstalled mapState: mapState
	overrideRedirect: overrideRedirect colormap: colormap 
	allEventMasks: allEventMasks yourEventMask: yourEventMask
	doNotPropagateMask: doNotPropagateMask
!

listInstalledColormaps
    | req numCmaps cmaps |
    req _ XRequest fixedCommand: 83.
    req ulong: id.
    req queueOn: server.
    
    server getReply.
    numCmaps _ server word.
    cmaps _ Array new: numCmaps.
    server skip: 22.
    1 to: numCmaps do: [ :i | cmaps at: i put: (Colormap idFrom: server) ].
    ^ListInstalledColormapsReply cmaps: cmaps
!

listProperties
    | numAtoms atoms |
    self justId: 21.
    
    server getReply.
    numAtoms _ server word.
    atoms _ Array new: numAtoms.
    server skip: 22.
    1 to: numAtoms do: [ :i | atoms at: i put: (Atom idFrom: server) ].
    ^ListPropertiesReply atoms: atoms
!


mapSubwindows
    self justId: 9
!

mapWindow
    self justId: 8
!

queryPointer
    | sameScreen root child rootX rootY winX winY mask |
    self justId: 38.
    
    sameScreen _ (server getReply) = 1.	"cheap bool conversion"
    root _ Window idFrom: server.
    child _ Window maybeIdFrom: server symbols: #(None).
    rootX _ server word.
    rootY _ server word.
    winX _ server word.
    winY _ server word.
    mask _ server uword "as events".
    server skip: 6.
    ^QueryPointerReply sameScreen: sameScreen root: root child: child
		       rootX: rootX rootY: rootY winX: winX winY: winY
		       mask: mask
!

queryTree
    | root parent numWin children |
    self justId: 15.
    
    server getReply.
    root _ Window idFrom: server.
    parent _ Window maybeIdFrom: server symbols: #(None).
    numWin _ server word.
    server skip: 14.
    children _ Array new: numWin.
    1 to: numWin do: [ :i | children at: i put: (Window idFrom: server) ].
    ^QueryTreeReply root: root parent: parent children: children
!

reparentWindow: parent x: x y: y
    | req |
    req _ XRequest fixedCommand: 7.
    req ulong: id; ulong: parent id; word: x; word: y.
    req queueOn: server.
!

rotateProperties: delta properties: properties
    | req |
    req _ XRequest command: 114.
    req ulong: id; word: properties size; word: delta.
    properties do: [ :prop | req ulong: prop id ].
    req queueOn: server.
!

translateCoordinates: srcWindow srcX: srcX srcY: srcY 
    | req sameScreen child dstX dstY |
    req _ XRequest fixedCommand: 40.
    req ulong: srcWindow id; ulong: id; word: srcX; word: srcY.
    req queueOn: server.

    sameScreen _ (server getReply) = 1.	"cheap bool conversion"
    child _ Window maybeIdFrom: server symbols: #(None).
    dstX _ server word.
    dstY _ server word.
    server skip: 16.
    ^TranslateCoordinatesReply sameScreen: sameScreen child: child
		       dstX: dstX dstY: dstY
!

unmapSubwindows
    self justId: 11
!

unmapWindow
    self justId: 10
! !



!Window methodsFor: 'useful extensions'!

name: aString
    self changeProperty: (Atom prop: #WmName on: server)
	 type: (Atom prop: #String on: server)
	 format: 8
	 mode: #Replace
	 data: aString.

!

iconName: aString
    self changeProperty: (Atom prop: #WmIconName on: server)
	 type: (Atom prop: #String on: server)
	 format: 8
	 mode: #Replace
	 data: aString.
! !

LoadStatus _ 'Window'!

