"======================================================================
|
|   Basic X protocol building blocks.
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     16 Feb 92	  created fall 90.
|
"
X subclass: #XProtocol
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''
  category: 'X protocol interface'
!

XProtocol comment: '###No comment yet'!


!XProtocol class methodsFor: 'subclass creation'!

defineAccessors: nameList forClass: class
    nameList do:
	[ :name | class compile: (name, '
    ^', name, '
') ]
!


defineModifiers: nameList forClass: class
    nameList do:
	[ :name | class compile: (name, ': ', name, 'XXX
    ', name, ' _ ', name, 'XXX', '
') ].
!

defineInstCreation: nameList forClass: class
    | selectorList |
    selectorList _ self selectorListFor: nameList.
    class class compile: (selectorList, '
    ^self new ', selectorList).

    " ### should make this category private"
    class compile: (selectorList, '
    ', (self initializationStringFor: nameList))
!



selectorListFor: instanceVariableNames
    | str |
    str _ WriteStream on: (String new: 0).
    instanceVariableNames do: 
	[ :name | str nextPutAll: name; nextPutAll: ': ';
		      nextPutAll: name; nextPutAll: 'XXX'; space ].
    ^str contents
!

initializationStringFor: instanceVariableNames
    | str |
    str _ WriteStream on: (String new: 0).
    instanceVariableNames do: 
	[ :name | str tab; nextPutAll: name; nextPutAll: ' _ ';
		      nextPutAll: name; nextPutAll: 'XXX.'; nl ].
    ^str contents
! !



!XProtocol class methodsFor: 'X utilities'!

map: value into: symbolArray
    | index |
    index _ symbolArray indexOf: value.
    index = 0
	ifTrue: [ ^self error: 'Could not map ', (value printString) ]
	ifFalse: [ ^index - 1 ]
!

maybeMap: valueOrInteger into: symbolArray
    valueOrInteger class == Integer
	ifTrue: [ ^valueOrInteger ]
	ifFalse: [ ^self map: valueOrInteger into: symbolArray ]
!	


maybeMapId: valueOrXObject into: symbolArray
    valueOrXObject class == Symbol
	ifTrue: [ ^self map: valueOrXObject into: symbolArray ]
	ifFalse: [ ^valueOrXObject id ]
!

mask: maskValue using: maskConstants
    | bitMask bit |
    maskValue class == Integer
	ifTrue: [ bitMask _ maskValue ]
	ifFalse: 
	    [ bitMask _ 0.
	      maskValue do: 
		  [ :maskName | bit _ self mapMask: maskName
					   into: maskConstants
					   and: #().
				bitMask _ bitMask bitOr: bit ]
	      ].
    ^bitMask
!

mask: maskValue using: maskConstants and: extraValueList
    | bitMask bit |
    "the extra value list here is of the form:
    ((symbol value) (symbol value) ...)"
    maskValue class == Integer
	ifTrue: [ bitMask _ maskValue ]
	ifFalse: 
	    [ 
	      maskValue class == Symbol
		  ifTrue: [ bitMask _ self mapMask: maskValue
					   into: maskConstants
					   and: extraValueList ]
		  ifFalse: [ bitMask _ 0.
			     maskValue do: 
				 [ :maskName | bit _ self mapMask: maskName 
							  into: maskConstants
							  and: extraValueList.
					       bitMask _ bitMask bitOr: bit ]
				 ]
	      ].
    ^bitMask
!

unmap: value into: symbolArray
    ^symbolArray at: value + 1
!

mapBool: aBool
    aBool			"expand mapping later"
	ifTrue: [ ^1 ]
	ifFalse: [ ^0 ]
! !	



!XProtocol class methodsFor: 'initialization'!

initialize
    "### These names could be put into some XProtocol pool dictionary, instead
     of polluting the global namespace"
    | names eventNames |
    eventNames _ #(KeyPress KeyRelease ButtonPress ButtonRelease EnterWindow
		  LeaveWindow PointerMotion PointerMotionHint
		  Button1Motion Button2Motion Button3Motion Button4Motion
		  Button5Motion ButtonMotion KeymapState Exposure
		  VisibilityChange StructureNotify ResizeRedirect
		  SubstructureNotify SubstructureRedirect FocusChange
		  PropertyChange ColormapChange OwnerGrabButton).
    self namesAndMasks: #Event from: eventNames.
    names _ self remove: #(KeyPress KeyRelease Exposure
			      VisibilityChange StructureNotify
			      ResizeRedirect SubstructureNotify
			      SubstructureRedirect FocusChange
			      PropertyChange ColormapChange
			      OwnerGrabButton)
	from: (eventNames copy).
    self namesAndMasks: #PointerEvent from: names.
    names _ self remove: #(EnterWindow LeaveWindow PointerMotionHint
			       KeymapState Exposure VisibilityChange
			       StructureNotify ResizeRedirect
			       SubstructureNotify SubstructureRedirect
			       FocusChange PropertyChange ColormapChange
			       OwnerGrabButton)
		 from: (eventNames copy).
    self namesAndMasks: #DeviceEvent from: names.
    names _ #(Shift Lock Control Mod1 Mod2 Mod3 Mod4 Mod5).
    self namesAndMasks: #KeyMask from: names.
    names _ names , #(Button1 Button2 Button3 Button4 Button5).
    self namesAndMasks: #KeyButMask from: names.
    names _ #(Forget NorthWest North NorthEast
		  West Center East SouthWest
		  South SouthEast Static).
    self namesAndMasks: #BitGravity from: names.
    names _ names copy.
    names at: 1 put: #Unmap.
    self namesAndMasks: #WinGravity from: names.
!

remove: eventNameList from: eventNames
    1 to: eventNames size do:
	[ :i | (eventNameList includes: (eventNames at: i))
		   ifTrue: [ eventNames at: i put: nil ] ].
    ^eventNames
	
!

namesAndMasks: baseName from: names
    | maskDict name |
    Smalltalk at: ((baseName, 'Names') asSymbol) put: names.
    maskDict _ Dictionary new.
    1 to: names size do:
	[ :i | name _ names at: i.
	       name notNil 
		   ifTrue: [ maskDict at: (name, 'Mask') asSymbol
				      put: (1 bitShift: (i - 1)) ]
		   ].
    Smalltalk at: ((baseName, 'MaskNames') asSymbol) put: maskDict
! !



!XProtocol class methodsFor: 'private'!

mapMask: maskName into: maskConstants and: extraValueList
    | index |
    index _ maskConstants indexOf: maskName.
    index > 0 ifTrue: [ ^1 bitShift: (index - 1) ].
    extraValueList do:
	[ :nameVal | (nameVal at: 1) == maskName
			 ifTrue: [ ^nameVal at: 2 ] ].
    ^self error: 'No value for bit named ', (maskName printString)
!!


XProtocol initialize!

LoadStatus _ 'XProtocol'!

FileStream fileIn: 'XError.st'!
FileStream fileIn: 'XEvent.st'!
FileStream fileIn: 'XClass.st'!
FileStream fileIn: 'XProtoResponses.st'!
FileStream fileIn: 'XRequest.st'!
FileStream fileIn: 'XServerObject.st'!

