"======================================================================
|
|   An early experimental STIX based browser prototype.
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     16 Feb 92	  created around christmas 91
|
"
Smalltalk at: #EventHandlers put: Dictionary new!
Smalltalk at: #ExitHandlerLoop put: nil!
Smalltalk at: #AccessorMethods put: Dictionary new!

FileStream fileIn: 'Slider.st'!
FileStream fileIn: 'LabelledButton.st'!
FileStream fileIn: 'WM.st'!

FileStream fileIn: 'WinEvents.st'!

Widget subclass: #Brow
       instanceVariableNames: 'background array gc font
                               fontInfo height yStart maxY slider 
                               winWidth winHeight curIndex xorGC'
       classVariableNames: 'ServerConn'
       poolDictionaries: ''
       category: 'Browsing experimental'
!

!Brow class methodsFor: 'Instance creation'!

on: array
    | inst server |
    server _ self newServer.
    inst _ self newWindow: server.
    inst init: array.
    ^inst
!

newWindow: s
    | root screen rootWin w winWidth winHeight |
    winWidth _ 700.
    winHeight _ 400.
    root _ s root.
    screen _ root screens at: 1.
    rootWin _ screen window.
    w _ Brow createWindow: rootWin class: #CopyFromParent
	       depth: 0
	       visual: #CopyFromParent x: 0 y: 0
	       width: winWidth height: winHeight borderWidth: 5 onServer: s
	       attrs: [ :r | r eventMask: #(Exposure ButtonPress 
						ButtonRelease
						"ResizeRedirect"
						Button1Motion
						StructureNotify);
				 backgroundPixel: screen blackPixel].
    w mapWindow.
    ^w
!

newServer
    | s | 
    ServerConn notNil ifTrue: [ServerConn close].
    ServerConn _ s _ Server connectToDisplay.
    ^s
! !


!Brow methodsFor: 'accessing'!

server
    ^server
!

fontInfo
    ^fontInfo
!

buttonHit: something
    'button hit' printNl.
    something inspect.
! !




!Brow methodsFor: 'private'!

init: array
    | normalHints root screen |
    root _ server root.
    screen _ root screens at: 1.

    winWidth _ 700.
    winHeight _ 400.
    "font _ Font openFont: '10x20' onServer: server."
    "font _ Font openFont: '-*-new century schoolbook-medium-r-normal--18-*-75-75-*-*-*-*' onServer: server."
    "font _ Font openFont: '-*-times-medium-r-normal--18-*-75-75-*-*-*-*' onServer: server."
    "font _ Font openFont: '-*-avantgarde-demi-r-normal--18-*-75-75-*-*-*-*' onServer: server."
    "font _ Font openFont: 'bembo-italic' onServer: server."
    "font _ Font openFont: 'helvetica' onServer: server."
    "font _ Font openFont: '-*-helvetica-medium-o-normal--18-*-75-75-*-*-*-*' onServer: server."
    font _ Font openFont: '-*-helvetica-medium-o-normal--18-*-*-*-*-*-*-*' onServer: server.
    RealFontInfo _ fontInfo _ font queryFont.
    gc _ GC createGC: self onServer: server 
	    args: [ :r | r font: font; 
			     foreground: screen whitePixel;
			     background: screen blackPixel ].
    height _ fontInfo fontAscent + fontInfo fontDescent.
    'height is ' print. height printNl.
    self initWin: array.

    xorGC _ GC createGC: self onServer: server
	     args: [ :r | r function: #Xor;
			      foreground: (screen whitePixel bitOr: screen blackPixel)";
			      background: screen blackPixel" ].
    LabelledButton createOn: self at: 10@40
		   notifying: self with: #buttonHit:
		   font: font
		   label: 'Foo on you'.

    normalHints _ WMSizeHints new flags: #(PResizeInc);
	widthInc: 30; heightInc: height.
    self changeProperty: (Atom prop: #WmNormalHints on: server)
	 type: (Atom prop: #WmSizeHints on: server)
	 format: 32
	 mode: #Replace data: normalHints.

!

blitIt
    self copyArea: background gc: gc srcX: 0 srcY: yStart * height
	 width: winWidth height: winHeight
	 dstX: 0 dstY: 0.
!

up
    | oldY |
    oldY _ yStart.
    yStart _ yStart - 1 max: 0.
    oldY ~= yStart
	ifTrue: [ curIndex notNil ifTrue: [ curIndex _ curIndex + 1. ].
		  self blitIt ].
!

upArrow
    self up
!

down
    | oldY |
    oldY _ yStart.
    yStart _ yStart + 1 min: maxY.
    oldY ~= yStart
	ifTrue: [ curIndex notNil ifTrue: [ curIndex _ curIndex - 1. ].
		  self blitIt ].
!

downArrow
    self down
!

initWin: lineArray
    | nLines totalHeight y bgc root screen points |
    root _ server root.
    screen _ root screens at: 1.
    yStart _ 0.
    array _ lineArray.
    totalHeight _ array size * height.
    background _ Pixmap createPixmap: self depth: 8
			width: winWidth height: totalHeight.
    maxY _ array size min: winHeight // height.
    bgc _ GC createGC: background onServer: server
	     args: [ :r | r foreground: screen blackPixel ].
    background polyFillRectangle: bgc
	       rectangles: (Array with: (0@0 corner: winWidth@totalHeight)).
    y _ fontInfo fontAscent.
    array do: 
	[ :elt | background imageText8: gc
		     x: 0 y: y string: elt.
		 y _ y + height ].
    slider _ Slider on: self notifying: self max: array size
		    min: 0 initial: 0.
    self setupHandlers.
!




sizeWindow: newWidth height: newHeight
    'size window' printNl.
    newWidth printNl.
    newHeight printNl.
    self configureWindow: 
	[ :r | r width: newWidth; height: (newHeight truncateTo: height) ].
!
    
    
trackPointer: e
    | index | 
    index _ e eventY // height.
    index ~= curIndex 
	ifTrue: [ self unhighlight.
		  (index between: 0 and: maxY)
		      ifTrue: [ curIndex _ index.
				self highlight. ]
		      ifFalse: [curIndex _  nil ] ].
!


highlight
    self setHighlight: true.
!

unhighlight
    curIndex notNil
	ifTrue: [ self setHighlight: false ].
!

setHighlight: desiredValue
    | rect |
    (curIndex between: 0 and: maxY) 
	ifTrue: [ rect _ 0@(curIndex * height) extent: winWidth@height.
		  self polyFillRectangle: xorGC
		       rectangles: (Array with: rect). ].
!


pressed: e
    self trackPointer: e
!

released: e
    self unhighlight.
    (curIndex notNil 
	and: [ curIndex between: 0 and: maxY ])
	ifTrue: [ (yStart + curIndex) printNl. ].
    curIndex _ nil.
!


setupHandlers
    | w h |
    self registerFor: ExposeEvent
	when: [ :e | e window == id ]
	do: [ :e | e count = 0
		       ifTrue: [ self blitIt ] ].
    self registerFor: ConfigureNotifyEvent
	when: [ :e | e event == id ]
	do: [ :e | w _ e width.
		   h _ e height.
		   (w ~= winWidth) | (h ~= winHeight)
		       ifTrue: [ winWidth _ w.
				 winHeight _ h.
				 slider resized.
				 maxY _ array size min: (winHeight // height).
				 self blitIt.
				 ]
		       ].
    self registerFor: ReparentNotifyEvent
	when: [ :e | e event == id ]
	do: [ :e | e inspect ].
    self registerFor: GraphicsExposureEvent
	when: [ :e | e drawable == id ]
	do: [ :e | true ].
    self registerFor: ButtonPressEvent
	when: [ :e | e event == id ]
	do: [ :e | self pressed: e ].
    self registerFor: ButtonReleaseEvent
	when: [ :e | e event == id ]
	do: [ :e | self released: e ].
    self registerFor: MotionNotifyEvent
	when: [ :e | e event == id ]
	do: [ :e | self trackPointer: e ].
    self registerFor: MapNotifyEvent
	when: [ :e | e event == id ]
	do: [ :e | e inspect ].
    self registerFor: UnmapNotifyEvent
	when: [ :e | e event == id ]
	do: [ :e | e inspect ].
    self  registerFor: ResizeRequestEvent
	when: [ :e | e window == id ]
	do: [ :e | self sizeWindow: e width height: e height ].
!

close
    self destroyWindow
! !


| b coll name |
    EventHandlers _ Dictionary new.

    Widget initialize.
    b _ Brow on: 
	(false ifTrue: [ coll _ OrderedCollection new.
			Object allSubclassesDo: 
			    [ :class | (name _ class name)
					   notNil ifTrue: [ coll add: name ].
				       ].
			coll add: Object name.
			coll asSortedCollection ]
	      ifFalse: [
	#(
	       FIRSTINTHELIST
	       Foo Object Bar Quem Juma
	       Foo Bar Object Quem Juma
	       Foo Bar Object Quem Juma
		   M5 The Multitronic Unit
		   Impressed Human Engrams
		   Dr Richard Daystrom
	       LASTINTHELIST
		   ) 
	    ]).
    b server processEvents.
    b close.
!

"Smalltalk quitPrimitive!

bogus
    b blitIt.
    ""50 timesRepeat: [ b up. ].
    50 timesRepeat: [ b down. ].""
    b server processEvents.
    b close.
!
"

"
| f |
    f _ FileStream open: 'brow.ps' mode: 'w'.
    Brow filePostscriptOutOn: f.
    f close
!
"
