"======================================================================
|
|   Show the color fountain in action.
|   Evaluate the expression in the comment at the bottom of the file to see it.
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     13 Sep 94	  Fixed white and black pixels to be derived from the
|			  screen values instead of hard coded.
|
| sbb	     28 Aug 94	  Added header stuff.
|
"

X subclass: #Thermo
       instanceVariableNames: 'mask invMask width height max min deltaW offset
			       value'
       classVariableNames: 'window'
       poolDictionaries: ''
       category: nil
! 

FileStream fileIn: 'WinEvents.st'!
FileStream fileIn: 'fountain.st'!


!Thermo class methodsFor: 'instance creation'!

new
    ^super new init
! !


!Thermo methodsFor: 'event handling'!

run
    window server processEvents.
!

stopRunning
    window server close.
    ExitHandlerLoop _ true.
! 

handleButtonPress: event
    event detail == 1
	ifTrue: [ ^self value: value + 1 ].
    event detail == 2
	ifTrue: [ ^self value: value - 1 ].
    event detail == 3
	ifTrue: [ self stopRunning ].
    
! !

!Thermo methodsFor: 'stuff'!

value: aValue
    | oldValue |
    oldValue _ value.
    value _ (aValue min: max) max: min.
    value ~= oldValue ifTrue: [  self drawContents ]
!

drawContents 
    | rect rects separator screen |
    rect _ 0@0 corner: width@0.
    rects _ Array with: rect.
    screen _ window server root screens at: 1.
    GC withGC: window
       args: [ :r | r clipMask: mask; foreground: screen blackPixel ]
       do: [ :gc | separator _ (value*(height+(width//2))//max).
		   window polyFillRectangle: gc
			  rectangles: (Array with: (0@0 corner: ((width+(2*deltaW))@separator) + offset)).
		   gc changeGC: [ :r | r foreground: screen whitePixel ].
		   window polyFillRectangle: gc
			  rectangles: 
			      (Array 
				   with: (0@(separator + offset)
					      corner: 
						  ((width+(2*deltaW))
						       @(height
							     +(width//2))) + offset)).
		   ].
!



drawWindow
    | fountain |
    fountain _ Fountain start: (RGB red: 16rFFFF green: 16rFFFF blue: 16rFFFF)
			 end: (RGB red: 16r8000 green: 16r8000 blue: 16rFFFF ).
"			 end: (RGB red: 0 green: 0 blue: 16rFFFF )."
    
    fountain _ Fountain start: (RGB red: 16rFFFF green: 16r8000 blue: 16r8000)
			 end: (RGB red: 16r8000 green: 16r8000 blue: 16rFFFF ).
"			 end: (RGB red: 0 green: 0 blue: 16rFFFF )."
    
    fountain _ Fountain start: (RGB red: 16r0 green: 16rFFFF blue: 16r8000)
			 end: (RGB red: 16rFFFF green: 16r8000 blue: 16rFFFF ).
"			 end: (RGB red: 16rFFFF green: 0 blue: 16r0 )."
    
    fountain drawOn: window.
    GC withGC: window
       args: [ :r | r lineWidth: 10; clipMask: invMask; foreground: 4 ]
       do: [ :gc | self drawInto: window gc: gc filled: false ].
    self drawContents.
! !






!Thermo methodsFor: 'private'!

init
    | s root screen rootWin cmap |
    window notNil
	ifTrue: [ window server close.
		  " ### Very clumsy "
		  EventHandlers _ Dictionary new. ].
    s _ Server connectToDisplay.
    root _ s root.
    screen _ root screens at: 1.
    rootWin _ screen window.

    window _ Window createWindow: rootWin class: #CopyFromParent
		    depth: 0
		    visual: #CopyFromParent x: 0 y: 0
		    width: 400 height: 400
		    borderWidth: 1 onServer: s
		    attrs: [ :r | r eventMask: #(Exposure
						     ButtonPress
						     );
				      backingStore: #Always ].

    
    width _ 20.
    deltaW _ 20.
    height _ 200.
    max _ 100.
    min _ 0.
    offset _ 100.
    value _ 50.
    self createMasks.

    self registerEventHandlers.    
    window mapWindow.


!

registerEventHandlers
    window registerFor: ExposeEvent
	   when: [ :e | e window == window id ]
	   do: [ :e | e count = 0 
			  ifTrue: [ 'exposed' printNl.
				    self drawWindow ] ].
    window registerFor: NoExposureEvent
	   when: [ :e | true ]
	   do: [ :e | ].
    window registerFor: ButtonPressEvent
	   when: [ :e | e event == window id ]
	   do: [ :e | self handleButtonPress: e ].
    
! 

drawCircleAt: aPoint radius: radius on: drawable gc: gc filled: filled
    | op origin corner arcs |
    origin _ aPoint - radius.
    corner _ aPoint + radius.
    arcs _ Array with: (Arc x: origin x y: origin y
			    width: radius * 2 height: radius * 2
			    angle1: 0 angle2: 360*64).
    filled 
	ifTrue: [ op _ #polyFillArc:arcs: ]
	ifFalse: [ op _ #polyArc:arcs: ].
    drawable perform: op with: gc with: arcs.
!

drawRectangle: rect on: drawable gc: gc filled: filled
    | op rects |
    rects _ Array with: rect.
    filled 
	ifTrue: [ op _ #polyFillRectangle:rectangles: ]
	ifFalse: [ op _ #polyRectangle:rectangles: ].
    drawable perform: op with: gc with: rects.
!

drawInto: drawable gc: gc filled: filled
    self drawCircleAt: ((width//2)@height + offset) radius: width//2 
	 on: drawable gc: gc filled: filled.
    self drawRectangle: (0@0 + offset corner: width@height + offset)
	 on: drawable gc: gc filled: filled.
    self drawCircleAt: ((width//2)@0 + offset) radius: (width//2)+deltaW 
	 on: drawable gc: gc filled: filled.
!

createMasks
    | pw ph  |
    pw _ 400. 
    ph _ 400.
    mask _ Pixmap createPixmap: window depth: 1 
		width: pw height: ph.
    GC withGC: mask 
       args: [ :r | r foreground: 1; background: 0; function: #Clear ]
       do: [ :gc | 
    
		   mask polyFillRectangle: gc rectangles:
		       (Array with: (0@0 corner: pw@ph)).
		   gc changeGC: [ :r | r function: #Copy ].
		   self drawInto: mask gc: gc filled: true.
		   invMask _ Pixmap createPixmap: window depth: 1 
				    width: pw height: ph.
		   gc changeGC: [ :r | r function: #CopyInverted ].
		   invMask copyArea: mask gc: gc
			   srcX: 0 srcY: 0 
			   width: pw height: ph
			   dstX: 0 dstY: 0.
		   ].

! !

"
| x |
    x _ Thermo new.
    x run.
!
"
